package de.ozgcloud.nachrichten.postfach.osiv2.config;

import org.apache.hc.client5.http.auth.AuthScope;
import org.apache.hc.client5.http.auth.UsernamePasswordCredentials;
import org.apache.hc.client5.http.impl.auth.BasicCredentialsProvider;
import org.apache.hc.client5.http.impl.classic.HttpClientBuilder;
import org.apache.hc.core5.http.HttpHost;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.http.client.ClientHttpRequestFactory;
import org.springframework.http.client.HttpComponentsClientHttpRequestFactory;
import org.springframework.http.converter.FormHttpMessageConverter;
import org.springframework.security.oauth2.client.AuthorizedClientServiceOAuth2AuthorizedClientManager;
import org.springframework.security.oauth2.client.InMemoryOAuth2AuthorizedClientService;
import org.springframework.security.oauth2.client.OAuth2AuthorizedClientProvider;
import org.springframework.security.oauth2.client.OAuth2AuthorizedClientProviderBuilder;
import org.springframework.security.oauth2.client.endpoint.RestClientClientCredentialsTokenResponseClient;
import org.springframework.security.oauth2.client.http.OAuth2ErrorResponseErrorHandler;
import org.springframework.security.oauth2.client.registration.ClientRegistration;
import org.springframework.security.oauth2.client.registration.ClientRegistrationRepository;
import org.springframework.security.oauth2.client.registration.InMemoryClientRegistrationRepository;
import org.springframework.security.oauth2.client.web.client.OAuth2ClientHttpRequestInterceptor;
import org.springframework.security.oauth2.core.AuthorizationGrantType;
import org.springframework.security.oauth2.core.ClientAuthenticationMethod;
import org.springframework.security.oauth2.core.http.converter.OAuth2AccessTokenResponseHttpMessageConverter;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.client.RestClient;

import de.ozgcloud.nachrichten.postfach.osiv2.gen.ApiClient;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.api.MailboxDirectoryApi;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.api.MessageExchangeApi;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.api.QuarantineApi;
import lombok.RequiredArgsConstructor;
import lombok.SneakyThrows;
import lombok.extern.log4j.Log4j2;

@Log4j2
@Configuration
@RequiredArgsConstructor
@ConditionalOnProperty(prefix = Osi2PostfachProperties.PREFIX, name = "enabled", havingValue = "true")
public class ApiClientConfiguration {

	private final Osi2PostfachProperties.AuthConfiguration authConfiguration;
	private final Osi2PostfachProperties.ApiConfiguration apiConfiguration;
	private final Osi2PostfachProperties.ProxyConfiguration proxyConfiguration;

	private static final String CLIENT_REGISTRATION_ID = "osi2";

	@Bean
	MessageExchangeApi messageExchangeApi(ApiClient apiClient) {
		return new MessageExchangeApi(apiClient);
	}

	@Bean
	QuarantineApi quarantineApi(ApiClient apiClient) {
		return new QuarantineApi(apiClient);
	}

	@Bean
	MailboxDirectoryApi mailboxDirectoryApi(ApiClient apiClient) {
		return new MailboxDirectoryApi(apiClient);
	}

	@Bean
	@SneakyThrows
	ApiClient apiClient() {
		var apiClient = new ApiClient(restClient());
		LOG.info("Setting api client base path to {}", apiConfiguration.getUrl());
		apiClient.setBasePath(apiConfiguration.getUrl());
		return apiClient;
	}

	private RestClient restClient() {
		var requestInterceptor = new OAuth2ClientHttpRequestInterceptor(authorizedClientManager());
		requestInterceptor.setClientRegistrationIdResolver(request -> CLIENT_REGISTRATION_ID);

		return defaultRestClientBuilder()
				.requestInterceptor(requestInterceptor)
				.build();
	}

	private RestClient.Builder defaultRestClientBuilder() {
		return RestClient.builder()
				.requestFactory(createProxyRequestFactory());
	}

	private ClientHttpRequestFactory createProxyRequestFactory() {
		var requestFactory = new HttpComponentsClientHttpRequestFactory();
		if (proxyConfiguration.isEnabled()) {
			LOG.info("Using proxy configuration: {}:{}", proxyConfiguration.getHost(), proxyConfiguration.getPort());
			requestFactory.setHttpClient(
					HttpClientBuilder.create()
							.setProxy(new HttpHost(proxyConfiguration.getHost(), proxyConfiguration.getPort()))
							.setDefaultCredentialsProvider(basicCredentialsProviderForProxy())
							.build()
			);
		}
		return requestFactory;
	}

	private BasicCredentialsProvider basicCredentialsProviderForProxy() {
		var credentialsProvider = new BasicCredentialsProvider();
		var username = proxyConfiguration.getUsername();
		var password = proxyConfiguration.getPassword();
		if (username != null && password != null) {
			LOG.info("Using proxy authentication with username {}", username);
			credentialsProvider.setCredentials(new AuthScope(proxyConfiguration.getHost(), proxyConfiguration.getPort()),
					new UsernamePasswordCredentials(username, password.toCharArray()));
		}
		return credentialsProvider;
	}

	private AuthorizedClientServiceOAuth2AuthorizedClientManager authorizedClientManager() {
		var clientRegistrations = clientRegistrationsRepository();
		var clientService = new InMemoryOAuth2AuthorizedClientService(
				clientRegistrations);
		var authorizedClientManager = new AuthorizedClientServiceOAuth2AuthorizedClientManager(
				clientRegistrations, clientService);

		authorizedClientManager.setAuthorizedClientProvider(authorizedClientProvider());

		return authorizedClientManager;
	}

	private ClientRegistrationRepository clientRegistrationsRepository() {
		return new InMemoryClientRegistrationRepository(osi2ClientRegistration());
	}

	private ClientRegistration osi2ClientRegistration() {
		LOG.info("Creating client registration with clientId: {}, tokenUri: {}, scope: {}, resource: {}", authConfiguration.getClientId(),
				authConfiguration.getTokenUri(), authConfiguration.getScope(), authConfiguration.getResource());
		return ClientRegistration.withRegistrationId(CLIENT_REGISTRATION_ID)
				.clientAuthenticationMethod(ClientAuthenticationMethod.CLIENT_SECRET_POST)
				.authorizationGrantType(AuthorizationGrantType.CLIENT_CREDENTIALS)
				.clientId(authConfiguration.getClientId())
				.clientSecret(authConfiguration.getClientSecret())
				.tokenUri(authConfiguration.getTokenUri())
				.scope(authConfiguration.getScope())
				.build();
	}

	private OAuth2AuthorizedClientProvider authorizedClientProvider() {
		return OAuth2AuthorizedClientProviderBuilder.builder()
				.clientCredentials(builder ->
						builder.accessTokenResponseClient(clientCredentialsTokenResponseClient())
				)
				.build();
	}

	private RestClientClientCredentialsTokenResponseClient clientCredentialsTokenResponseClient() {
		var client = new RestClientClientCredentialsTokenResponseClient();
		configureClientCredentialsRestClient(client);
		configureParametersForTokenRequests(client);
		return client;
	}

	private void configureClientCredentialsRestClient(RestClientClientCredentialsTokenResponseClient client) {
		client.setRestClient(defaultRestClientBuilder()
				.messageConverters(messageConverters -> {
					messageConverters.clear();
					messageConverters.add(new FormHttpMessageConverter());
					messageConverters.add(new OAuth2AccessTokenResponseHttpMessageConverter());
				})
				.defaultStatusHandler(new OAuth2ErrorResponseErrorHandler())
				.build());
	}

	private void configureParametersForTokenRequests(RestClientClientCredentialsTokenResponseClient client) {
		client.addParametersConverter(source -> {
			MultiValueMap<String, String> parameters = new LinkedMultiValueMap<>();
			// Pass a resource indicator parameter https://datatracker.ietf.org/doc/html/rfc8707
			parameters.add("resource", authConfiguration.getResource());
			return parameters;
		});
	}

}
