package de.ozgcloud.nachrichten.postfach.osiv2.transfer;

import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.function.Function;
import java.util.stream.Collectors;

import org.mapstruct.Mapper;
import org.mapstruct.Mapping;
import org.mapstruct.Named;
import org.mapstruct.ReportingPolicy;

import de.ozgcloud.nachrichten.postfach.PostfachAddress;
import de.ozgcloud.nachrichten.postfach.PostfachAddressIdentifier;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.StringBasedIdentifier;
import de.ozgcloud.nachrichten.postfach.osiv2.OsiPostfachRemoteService;
import de.ozgcloud.nachrichten.postfach.osiv2.model.AttachmentInfo;
import de.ozgcloud.nachrichten.postfach.osiv2.model.Osi2Attachment;
import de.ozgcloud.nachrichten.postfach.osiv2.model.Osi2Message;

@Mapper(unmappedTargetPolicy = ReportingPolicy.ERROR)
public interface Osi2MessageMapper {
	String POSTFACH_ADDRESS_VERSION = "2.0";
	int POSTFACH_ADDRESS_TYPE = 2;

	@Mapping(target = "id", ignore = true)
	@Mapping(target = "referencedNachricht", ignore = true)

	@Mapping(target = "createdBy", ignore = true)

	@Mapping(target = "sentAt", ignore = true)
	@Mapping(target = "sentSuccessful", ignore = true)
	@Mapping(target = "messageCode", ignore = true)

	@Mapping(target = "postfachAddress", source = "osi2Message.mailboxId", qualifiedByName = "createPostfachAddressByMailboxId")
	@Mapping(target = "attachments", source = "attachmentIds")
	@Mapping(target = "direction", constant = "IN")
	PostfachNachricht mapPostfachNachricht(Osi2Message osi2Message, List<String> attachmentIds);

	@Named("createPostfachAddressByMailboxId")
	default PostfachAddress createPostfachAddressByMailboxId(String mailBoxId) {
		return PostfachAddress.builder()
				.type(POSTFACH_ADDRESS_TYPE)
				.version(POSTFACH_ADDRESS_VERSION)
				.identifier(StringBasedIdentifier.builder()
						// Note: this should be a nameIdentifier for sending but is a mailboxGuid/messageBox.
						// Thus, the PostfachAddress in this mapped response can not be used for sending
						.postfachId(mailBoxId)
						.build())
				.serviceKontoType(OsiPostfachRemoteService.POSTFACH_TYPE_OSI)
				.build();

	}

	default String getNameIdentifier(PostfachNachricht nachricht) {
		return Optional.ofNullable(nachricht.getPostfachAddress())
				.map(PostfachAddress::getIdentifier)
				.map(PostfachAddressIdentifier::getStringRepresentation)
				.orElseThrow(() -> new IllegalArgumentException("Missing nameIdentifier!"));
	}

	@Mapping(target = "mailboxId", source = "mailboxId")
	@Mapping(target = "attachments", expression = "java( mapAttachmentInfosInOriginalOrder(nachricht, uploadAttachments) )")
	Osi2Message mapOsi2Message(PostfachNachricht nachricht, String mailboxId, List<Osi2Attachment> uploadAttachments);

	default List<AttachmentInfo> mapAttachmentInfosInOriginalOrder(PostfachNachricht nachricht, List<Osi2Attachment> files) {
		var filesById = associateUploadWithAttachmentId(files);
		return nachricht.getAttachments()
				.stream()
				.map(fileId -> Objects.requireNonNull(
						filesById.get(fileId),
						"Expect all attachmentIds are uploaded!"
				))
				.map(this::mapAttachmentInfo)
				.toList();
	}

	@Mapping(target = "contentType", source = "file.contentType")
	@Mapping(target = "name", source = "file.name")
	@Mapping(target = "size", source = "file.size")
	AttachmentInfo mapAttachmentInfo(Osi2Attachment file);

	default Map<String, Osi2Attachment> associateUploadWithAttachmentId(List<Osi2Attachment> uploads) {
		return uploads.stream()
				.filter(upload -> upload.file() != null && upload.file().getId() != null)
				.collect(Collectors.toMap(upload -> upload.file().getId().toString(), Function.identity()));
	}
}
