package de.ozgcloud.nachrichten.postfach.osiv2.transfer;

import static de.ozgcloud.nachrichten.postfach.osiv2.transfer.WaitUtil.*;

import java.io.IOException;
import java.io.InputStream;
import java.time.Duration;
import java.util.Comparator;
import java.util.List;
import java.util.stream.LongStream;
import java.util.stream.Stream;

import de.ozgcloud.apilib.file.OzgCloudFile;
import de.ozgcloud.nachrichten.postfach.osiv2.ServiceIfOsi2Enabled;
import de.ozgcloud.nachrichten.postfach.osiv2.attachment.Osi2AttachmentFileService;
import de.ozgcloud.nachrichten.postfach.osiv2.exception.Osi2RuntimeException;
import de.ozgcloud.nachrichten.postfach.osiv2.exception.Osi2UploadException;
import de.ozgcloud.nachrichten.postfach.osiv2.model.FileChunkInfo;
import de.ozgcloud.nachrichten.postfach.osiv2.model.Osi2Attachment;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Log4j2
@ServiceIfOsi2Enabled
@RequiredArgsConstructor
public class Osi2QuarantineService {
	private final PostfachApiFacadeService postfachApiFacadeService;
	private final Osi2AttachmentFileService binaryFileService;

	static final Duration POLLING_INTERVAL = Duration.ofSeconds(1);
	static final Duration POLLING_TIMEOUT = Duration.ofMinutes(5);

	public List<Osi2Attachment> uploadAttachments(List<String> attachmentIds) {
		return uploadFiles(binaryFileService.getFileMetadataOfIds(attachmentIds));
	}

	List<Osi2Attachment> uploadFiles(List<OzgCloudFile> ozgCloudFiles) {
		var sortedUploadFiles = deriveSortedUploadFiles(ozgCloudFiles);
		try {
			tryUploadSortedFiles(sortedUploadFiles);
			return sortedUploadFiles;
		} catch (RuntimeException e) {
			deleteAttachments(sortedUploadFiles);
			throw e;
		}
	}

	List<Osi2Attachment> deriveSortedUploadFiles(List<OzgCloudFile> ozgCloudFiles) {
		return ozgCloudFiles.stream()
				.sorted(Comparator.comparing(OzgCloudFile::getSize).reversed())
				.map(Osi2Attachment::from)
				.toList();
	}

	void tryUploadSortedFiles(List<Osi2Attachment> sortedUploadFiles) {
		uploadFilesToQuarantine(sortedUploadFiles);
		waitForVirusScan(sortedUploadFiles);
	}

	void uploadFilesToQuarantine(List<Osi2Attachment> uploads) {
		uploads.forEach(this::uploadFileToQuarantine);
	}

	void uploadFileToQuarantine(Osi2Attachment upload) {
		try (var fileInputStream = binaryFileService.downloadFileContent(upload.file().getId().toString())) {
			uploadInputStreamToQuarantine(upload, fileInputStream);
		} catch (IOException e) {
			throw new Osi2RuntimeException("Failed to close upload input stream!", e);
		}
	}

	void uploadInputStreamToQuarantine(Osi2Attachment upload, InputStream fileInputStream) {
		streamFileChunkInfos(upload)
				.forEachOrdered(chunkInfo ->
						postfachApiFacadeService.uploadChunk(
								chunkInfo,
								chunkInfo.createUploadResource(fileInputStream)
						)
				);
	}

	Stream<FileChunkInfo> streamFileChunkInfos(Osi2Attachment upload) {
		return LongStream.range(0, upload.numberOfChunks() + 1)
				.mapToObj(chunkIndex -> buildFileChunkInfo(upload, chunkIndex));
	}

	private FileChunkInfo buildFileChunkInfo(Osi2Attachment upload, long chunkIndex) {
		return FileChunkInfo.builder()
				.upload(upload)
				.chunkIndex(chunkIndex)
				.build();
	}

	void waitForVirusScan(List<Osi2Attachment> osi2FileMetadata) {
		if (!waitUntil(() -> checkVirusScanCompleted(osi2FileMetadata), POLLING_INTERVAL, POLLING_TIMEOUT)) {
			throw new Osi2RuntimeException("Expect the scan to complete after %d seconds!".formatted(POLLING_TIMEOUT.getSeconds()));
		}
	}

	synchronized boolean checkVirusScanCompleted(List<Osi2Attachment> osi2FileMetadata) {
		return osi2FileMetadata.stream()
				.allMatch(this::checkOneVirusScanCompleted);
	}

	boolean checkOneVirusScanCompleted(Osi2Attachment osi2FileMetadata) {
		try {
			return postfachApiFacadeService.checkUploadSuccessful(osi2FileMetadata.guid());
		} catch (Osi2UploadException e) {
			throw new Osi2RuntimeException("%s failed!".formatted(osi2FileMetadata.getLoggableString()), e);
		}
	}

	void deleteAttachments(List<Osi2Attachment> uploads) {
		uploads.stream()
				.map(Osi2Attachment::guid)
				.forEach(postfachApiFacadeService::deleteFileUpload);
	}

}
