package de.ozgcloud.nachrichten.postfach.osiv2.transfer;

import java.time.OffsetDateTime;
import java.time.ZonedDateTime;
import java.util.Collection;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.UUID;
import java.util.stream.Stream;

import jakarta.validation.constraints.NotNull;

import org.mapstruct.Mapper;
import org.mapstruct.Mapping;
import org.mapstruct.Named;
import org.mapstruct.ReportingPolicy;

import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.osiv2.exception.Osi2MailboxNotFoundException;
import de.ozgcloud.nachrichten.postfach.osiv2.exception.Osi2RuntimeException;
import de.ozgcloud.nachrichten.postfach.osiv2.exception.Osi2UploadException;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.DTOMailboxScopeData;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.MddiMailboxIndexData;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.MessageExchangeReceiveMessage;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.MessageExchangeReceiveMessagesResponse;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.QuarantineFileResult;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.QuarantineStatus;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.ResponsesMailboxDirectoryResponse;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.V1ReplyBehavior;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.V1ReplyFiles;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.V1ReplyMessage;
import de.ozgcloud.nachrichten.postfach.osiv2.model.AttachmentInfo;
import de.ozgcloud.nachrichten.postfach.osiv2.model.Osi2Message;

@Mapper(unmappedTargetPolicy = ReportingPolicy.ERROR, imports = { Osi2HtmlDocument.class })
public interface Osi2ResponseMapper {

	@Mapping(target = "vorgangId", source = "sequencenumber")
	@Mapping(target = "mailboxId", source = "messageBox")
	@Mapping(target = "messageId", source = "guid")
	@Mapping(target = "createdAt", source = "responseTime", qualifiedByName = "mapOffsetDateTimeToZoned")
	@Mapping(target = "subject", source = "subject")
	@Mapping(target = "mailBody", source = ".", qualifiedByName = "mapMailBody")
	@Mapping(target = "replyOption", source = "replyAction")
	@Mapping(target = "attachments", source = "files", defaultExpression = "java(List.of())")
	Osi2Message toMessage(V1ReplyMessage message);

	@Mapping(target = "contentType", source = "mimeType")
	AttachmentInfo toAttachmentInfo(V1ReplyFiles attachment);

	default String mapNullToEmpty(String value) {
		return value == null ? "" : value;
	}

	@Named("mapOffsetDateTimeToZoned")
	default ZonedDateTime mapOffsetDateTimeToZoned(OffsetDateTime offsetDateTime) {
		return offsetDateTime == null ? null : offsetDateTime.toZonedDateTime();
	}

	@Named("mapMailBody")
	default String mapMailBody(V1ReplyMessage message) {
		var body = mapNullToEmpty(message.getBody());
		return Boolean.TRUE.equals(message.getIsHtml())
				? Osi2HtmlDocument.renderToPlainText(body)
				: body;
	}

	default PostfachNachricht.ReplyOption mapReplyAction(V1ReplyBehavior replyOption) {
		return replyOption == null
				? PostfachNachricht.ReplyOption.FORBIDDEN
				: switch (replyOption) {
					case REPLYFORBIDDEN -> PostfachNachricht.ReplyOption.FORBIDDEN;
					case REPLYPOSSIBLE -> PostfachNachricht.ReplyOption.POSSIBLE;
					case REPLYMANDATORY -> PostfachNachricht.ReplyOption.MANDATORY;
				};
	}

	default List<String> toMessageIds(MessageExchangeReceiveMessagesResponse response) {
		if (response == null) {
			throw new Osi2RuntimeException("Expect non empty response!", null);
		}

		return Optional.ofNullable(response.getMessages())
				.stream()
				.flatMap(Collection::stream)
				.map(MessageExchangeReceiveMessage::getGuid)
				.filter(Objects::nonNull)
				.map(UUID::toString)
				.toList();
	}

	default boolean isSafe(QuarantineStatus uploadStatus) throws Osi2UploadException {
		return switch (uploadStatus) {
			case NONE, UNSAFE, CORRUPT, MISSING -> throw new Osi2UploadException(uploadStatus);
			case LEGACY, COMMITED -> false;
			case SAFE -> true;
		};
	}

	default void checkChunkUploadSuccess(QuarantineFileResult quarantineFileResult) {
		if (!Optional.ofNullable(quarantineFileResult.getError()).map(String::isBlank).orElse(true)) {
			throw new Osi2RuntimeException(
					"Chunk-Upload of file %s failed: %s".formatted(quarantineFileResult.getFileUid(), quarantineFileResult.getError()), null);
		}
	}

	default Optional<String> extractMailboxIdByTenantOrFirst(@NotNull List<MddiMailboxIndexData> mailBoxes, String tenant) {
		return mailBoxes.stream()
				.filter(mailbox -> tenant.equals(mailbox.getTenant()))
				.findFirst()
				.or(() -> mailBoxes.stream().findFirst())
				.map(MddiMailboxIndexData::getMailboxguid)
				.map(UUID::toString);
	}

	default String getMailboxIdByTenantOrFirst(ResponsesMailboxDirectoryResponse responsesMailboxDirectoryResponse, String tenant) {
		return Optional.ofNullable(responsesMailboxDirectoryResponse)
				.map(ResponsesMailboxDirectoryResponse::getMailBoxIndex)
				.map(Collection::stream)
				.flatMap(Stream::findFirst)
				.map(DTOMailboxScopeData::getMemberscope)
				.flatMap(mailBoxes -> extractMailboxIdByTenantOrFirst(mailBoxes, tenant))
				.orElseThrow(() -> new Osi2MailboxNotFoundException("Empty mailbox lookup response! %s".formatted(
						responsesMailboxDirectoryResponse != null ? responsesMailboxDirectoryResponse.toString() : "(response null)")));
	}
}
