package de.ozgcloud.nachrichten.postfach.osiv2.transfer;

import static de.ozgcloud.nachrichten.postfach.osiv2.transfer.Osi2RequestMapper.*;

import java.util.List;
import java.util.UUID;

import org.springframework.core.io.AbstractResource;
import org.springframework.core.io.Resource;

import de.ozgcloud.nachrichten.postfach.osiv2.ServiceIfOsi2Enabled;
import de.ozgcloud.nachrichten.postfach.osiv2.config.Osi2PostfachProperties;
import de.ozgcloud.nachrichten.postfach.osiv2.exception.Osi2UploadException;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.api.MailboxDirectoryApi;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.api.MessageExchangeApi;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.api.QuarantineApi;
import de.ozgcloud.nachrichten.postfach.osiv2.model.FileChunkInfo;
import de.ozgcloud.nachrichten.postfach.osiv2.model.Osi2Message;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Log4j2
@ServiceIfOsi2Enabled
@RequiredArgsConstructor
public class PostfachApiFacadeService {

	private final MessageExchangeApi messageExchangeApi;
	private final QuarantineApi quarantineApi;
	private final MailboxDirectoryApi mailboxDirectoryApi;
	private final Osi2RequestMapper requestMapper;
	private final Osi2ResponseMapper responseMapper;
	private final Osi2PostfachProperties.ApiConfiguration apiConfiguration;

	public void sendMessage(Osi2Message message) {
		messageExchangeApi.sendMessage(
				message.mailboxId(),
				requestMapper.mapOutSendMessageRequestV2(message)
		);
	}

	public <T extends AbstractResource> void uploadChunk(FileChunkInfo chunkInfo, T chunkResource) {
		responseMapper.checkChunkUploadSuccess(quarantineApi.uploadChunk(
				requestMapper.mapDomainChunkMetaData(chunkInfo),
				apiConfiguration.getTenant(),
				apiConfiguration.getNameIdentifier(),
				chunkResource
		));
	}

	public boolean checkUploadSuccessful(final String messageId) throws Osi2UploadException {
		return responseMapper.isSafe(
				quarantineApi.getUploadStatus(UUID.fromString(messageId))
		);
	}

	public List<String> fetchPendingMessageIds() {
		return responseMapper.toMessageIds(
				messageExchangeApi.receiveMessages(MAX_NUMBER_RECEIVED_MESSAGES, 0)
		);
	}

	public Osi2Message fetchMessageById(final String messageId) {
		var messageReply = messageExchangeApi.getMessage(UUID.fromString(messageId));
		return responseMapper.toMessage(messageReply);
	}

	public void deleteMessage(final String messageId) {
		messageExchangeApi.deleteMessage(UUID.fromString(messageId));
	}

	public void deleteFileUpload(String fileUploadId) {
		quarantineApi.deleteUpload(UUID.fromString(fileUploadId));
	}

	public Resource downloadAttachment(String messageGuid, String attachmentGuid) {
		return messageExchangeApi.getMessageAttachment(UUID.fromString(messageGuid), UUID.fromString(attachmentGuid));
	}

	public String lookupMailboxId(String nameIdentifier) {
		return responseMapper.getMailboxIdByTenantOrFirst(
				mailboxDirectoryApi.lookupMailboxIds(List.of(UUID.fromString(nameIdentifier))),
				apiConfiguration.getTenant()
		);
	}
}
