package de.ozgcloud.nachrichten.postfach.osiv2;

import static com.github.tomakehurst.wiremock.client.WireMock.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.config.Osi2PostfachProperties.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.factory.JwtFactory.*;
import static org.assertj.core.api.Assertions.*;

import java.util.List;
import java.util.UUID;
import java.util.function.Function;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.RegisterExtension;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.DynamicPropertyRegistry;
import org.springframework.test.context.DynamicPropertySource;
import org.springframework.test.context.TestPropertySource;

import com.github.tomakehurst.wiremock.WireMockServer;
import com.github.tomakehurst.wiremock.client.ResponseDefinitionBuilder;
import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.nachrichten.postfach.PostfachMessageCode;
import de.ozgcloud.nachrichten.postfach.osiv2.attachment.Osi2AttachmentFileService;
import de.ozgcloud.nachrichten.postfach.osiv2.exception.Osi2PostfachException;
import de.ozgcloud.nachrichten.postfach.osiv2.extension.AttachmentExampleUploadUtil;
import de.ozgcloud.nachrichten.postfach.osiv2.extension.Jwt;
import de.ozgcloud.nachrichten.postfach.osiv2.extension.OsiMockServerExtension;
import de.ozgcloud.nachrichten.postfach.osiv2.extension.VorgangManagerServerExtension;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.JsonUtil;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.MessageExchangeReceiveMessagesResponseTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.MessageExchangeSendMessageResponseTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.PostfachNachrichtTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.ResponsesMailboxDirectoryResponseTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.V1ReplyFilesTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.V1ReplyMessageTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.QuarantineFileResult;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.QuarantineStatus;
import lombok.SneakyThrows;

@SpringBootTest(classes = TestApplication.class)
@ActiveProfiles({ "stage", "itcase" })
@TestPropertySource(properties = {
		"ozgcloud.osiv2.proxy.enabled=false",
})
class OsiPostfachRemoteServiceITCase {

	@RegisterExtension
	static final OsiMockServerExtension OSI_MOCK_SERVER_EXTENSION = new OsiMockServerExtension();

	@RegisterExtension
	static final VorgangManagerServerExtension VORGANG_MANAGER_SERVER_EXTENSION = new VorgangManagerServerExtension();

	@Autowired
	protected OsiPostfachRemoteService osiPostfachRemoteService;
	@Autowired
	protected Osi2AttachmentFileService osi2AttachmentFileService;

	@DynamicPropertySource
	static void dynamicProperties(DynamicPropertyRegistry registry) {
		registry.add("ozgcloud.osiv2.auth.token-uri", OSI_MOCK_SERVER_EXTENSION::getAccessTokenUrl);
		registry.add("ozgcloud.osiv2.auth.scope", () -> CLIENT_SCOPES);
		registry.add("ozgcloud.osiv2.auth.client-id", () -> CLIENT_ID);
		registry.add("ozgcloud.osiv2.api.url", OSI_MOCK_SERVER_EXTENSION::getPostfachFacadeUrl);
		registry.add("ozgcloud.osiv2.auth.resource", () -> RESOURCE_URN);
		registry.add("grpc.client." + GRPC_FILE_MANAGER_NAME + ".address", VORGANG_MANAGER_SERVER_EXTENSION::getVorgangManagerAddress);
		registry.add("grpc.client." + GRPC_FILE_MANAGER_NAME + ".negotiationType", () -> "PLAINTEXT");
	}

	protected WireMockServer postfachFacadeMockServer;

	@BeforeEach
	@SneakyThrows
	public void setup() {
		postfachFacadeMockServer = OSI_MOCK_SERVER_EXTENSION.getPostfachFacadeMockServer();
	}

	@DisplayName("should send request with jwt")
	@Test
	@SneakyThrows
	void shouldSendRequestWithJwt() {
		var postfachNachricht = PostfachNachrichtTestFactory.create();
		mockLookupResponse();
		mockSendResponse();

		osiPostfachRemoteService.sendMessage(postfachNachricht);

		var requests = postfachFacadeMockServer.findAll(
				postRequestedFor(urlPathTemplate("/MessageExchange/v1/Send/{mailboxId}")));
		assertThat(requests).hasSize(1);
		var jwt = Jwt.parseAuthHeaderValue(requests.getFirst().getHeader("Authorization"));
		assertThat(jwt.body().read("$.client_id", String.class)).isEqualTo(CLIENT_ID);
		assertThat(jwt.body().read("$.aud", String.class)).isEqualTo(RESOURCE_URN);
	}

	@DisplayName("should send message with attachment")
	@Test
	void shouldSendMessageWithAttachment() {
		var textFileId = AttachmentExampleUploadUtil.uploadTextFile(osi2AttachmentFileService);
		postfachFacadeMockServer.stubFor(post(urlPathTemplate("/Quarantine/v1/Upload/Chunked"))
				.willReturn(okJsonObj(QuarantineFileResult.builder()
						.success(true)
						.build()))
		);
		postfachFacadeMockServer.stubFor(get(urlPathTemplate("/Quarantine/v1/Upload/{messageGuid}"))
				.willReturn(okJsonObj(QuarantineStatus.SAFE))
		);
		var postfachNachrichtWithAttachment = PostfachNachrichtTestFactory.createBuilder()
				.attachments(List.of(textFileId))
				.build();
		mockLookupResponse();
		mockSendResponse();

		osiPostfachRemoteService.sendMessage(postfachNachrichtWithAttachment);

		var requests = postfachFacadeMockServer.findAll(
				postRequestedFor(urlPathTemplate("/Quarantine/v1/Upload/Chunked")));
		assertThat(requests).hasSize(2);
		Function<Integer, byte[]> requestBodyBytes = requestIndex -> requests.get(requestIndex).getPart("file").getBody().asBytes();
		// should send file in one chunk
		assertThat(requestBodyBytes.apply(0)).isEqualTo(AttachmentExampleUploadUtil.EXAMPLE_TEXT_DATA);
		// should send empty chunk to complete transfer
		assertThat(requestBodyBytes.apply(1)).isEmpty();
		postfachFacadeMockServer.verify(
				exactly(1),
				getRequestedFor(urlPathTemplate("/Quarantine/v1/Upload/{messageGuid}"))
		);
		postfachFacadeMockServer.verify(
				exactly(1),
				postRequestedFor(urlPathTemplate("/MessageExchange/v1/Send/{mailboxId}"))
		);
		postfachFacadeMockServer.verify(
				exactly(1),
				postRequestedFor(urlPathTemplate("/MailboxDirectory/v1/Lookup"))
		);
	}

	@DisplayName("should throw postfach exception with connection error code")
	@Test
	void shouldThrowPostfachExceptionWithConnectionErrorCode() {
		postfachFacadeMockServer.stop();

		var postfachNachricht = PostfachNachrichtTestFactory.create();

		assertThatThrownBy(() -> osiPostfachRemoteService.sendMessage(postfachNachricht))
				.isInstanceOf(Osi2PostfachException.class)
				.hasFieldOrPropertyWithValue("messageCode", PostfachMessageCode.SERVER_CONNECTION_FAILED_MESSAGE_CODE);
	}

	@DisplayName("should throw postfach exception with unknown postfach id error code")
	@Test
	void shouldThrowPostfachExceptionWithUnknownPostfachIdErrorCode() {
		postfachFacadeMockServer.stubFor(post(urlPathTemplate("/MessageExchange/v1/Send/{mailboxId}"))
				.willReturn(notFound())
		);
		var postfachNachricht = PostfachNachrichtTestFactory.create();

		assertThatThrownBy(() -> osiPostfachRemoteService.sendMessage(postfachNachricht))
				.isInstanceOf(Osi2PostfachException.class)
				.hasFieldOrPropertyWithValue("messageCode", PostfachMessageCode.SEND_FAILED_UNKNOWN_POSTFACH_ID_MESSAGE_CODE);
	}

	@DisplayName("should delete attachments on upload exception")
	@Test
	void shouldDeleteAttachmentsOnUploadException() {
		var textFileId = AttachmentExampleUploadUtil.uploadTextFile(osi2AttachmentFileService);
		postfachFacadeMockServer.stubFor(post(urlPathTemplate("/Quarantine/v1/Upload/Chunked"))
				.willReturn(okJsonObj(QuarantineFileResult.builder()
						.success(false)
						.error("Upload failure")
						.build()))
		);
		postfachFacadeMockServer.stubFor(delete(urlPathTemplate("/Quarantine/v1/Upload/{messageGuid}"))
				.willReturn(ok())
		);
		var postfachNachrichtWithAttachment = PostfachNachrichtTestFactory.createBuilder()
				.attachments(List.of(textFileId))
				.build();
		mockLookupResponse();

		try {
			osiPostfachRemoteService.sendMessage(postfachNachrichtWithAttachment);
		} catch (Osi2PostfachException e) {
			// ignore
		}

		postfachFacadeMockServer.verify(
				exactly(1),
				deleteRequestedFor(urlPathTemplate("/Quarantine/v1/Upload/{messageGuid}"))
		);
	}

	@DisplayName("should delete attachments on scan exception")
	@Test
	void shouldDeleteAttachmentsOnScanException() {
		var textFileId = AttachmentExampleUploadUtil.uploadTextFile(osi2AttachmentFileService);
		postfachFacadeMockServer.stubFor(post(urlPathTemplate("/Quarantine/v1/Upload/Chunked"))
				.willReturn(okJsonObj(QuarantineFileResult.builder()
						.success(true)
						.build()))
		);
		postfachFacadeMockServer.stubFor(get(urlPathTemplate("/Quarantine/v1/Upload/{messageGuid}"))
				.willReturn(okJsonObj(QuarantineStatus.UNSAFE))
		);
		postfachFacadeMockServer.stubFor(delete(urlPathTemplate("/Quarantine/v1/Upload/{messageGuid}"))
				.willReturn(ok())
		);
		var postfachNachrichtWithAttachment = PostfachNachrichtTestFactory.createBuilder()
				.attachments(List.of(textFileId))
				.build();
		mockLookupResponse();

		try {
			osiPostfachRemoteService.sendMessage(postfachNachrichtWithAttachment);
		} catch (Osi2PostfachException e) {
			// ignore
		}

		postfachFacadeMockServer.verify(
				exactly(1),
				deleteRequestedFor(urlPathTemplate("/Quarantine/v1/Upload/{messageGuid}"))
		);
	}

	private void mockLookupResponse() {
		// Stub mailboxids lookup response (MessageExchangeApi::lookupMailboxIds)
		postfachFacadeMockServer.stubFor(post(urlPathTemplate("/MailboxDirectory/v1/Lookup"))
				.willReturn(
						okJsonObj(
								ResponsesMailboxDirectoryResponseTestFactory.create()
						)
				)
		);
	}

	private void mockSendResponse() {
		// Stub message send response (MessageExchangeApi::sendMessage)
		postfachFacadeMockServer.stubFor(post(urlPathTemplate("/MessageExchange/v1/Send/{mailboxId}"))
				.willReturn(
						okJsonObj(
								MessageExchangeSendMessageResponseTestFactory.create()
										.messageId(UUID.randomUUID())
						)
				)
		);
	}

	@DisplayName("should receive one messages")
	@Test
	@SneakyThrows
	void shouldReceiveMessages() {
		mockPostfachMessageAndResponse("00000000-0000-0000-0000-000000000000");

		var messageList = osiPostfachRemoteService.getAllMessages().toList();

		assertThat(messageList).hasSize(1);
		var firstMessage = messageList.getFirst();
		assertThat(firstMessage.getAttachments()).hasSize(1);
		var fileMetadata = osi2AttachmentFileService.getFileMetadataOfIds(firstMessage.getAttachments());
		var firstAttachment = fileMetadata.getFirst();
		assertThat(firstAttachment.getName()).isNotBlank();
		assertThat(firstAttachment.getContentType()).isEqualTo("text/plain");
	}

	@DisplayName("should receive two messages")
	@Test
	@SneakyThrows
	void shouldReceiveTwoMessages() {
		mockPostfachMessageAndResponse("00000000-0000-0000-0000-000000000000", "00000000-0000-0000-0000-000000000001");

		var messageList = osiPostfachRemoteService.getAllMessages().toList();

		assertThat(messageList).hasSize(2);
	}

	private void mockPostfachMessageAndResponse(final String... messageGuids) {
		// Stub message listing response (MessageExchangeApi::receiveMessages)
		postfachFacadeMockServer.stubFor(get(urlPathEqualTo("/MessageExchange/v1/Receive"))
				.withQueryParam("take", equalTo("100"))
				.withQueryParam("skip", equalTo("0"))
				.willReturn(okJsonObj(MessageExchangeReceiveMessagesResponseTestFactory.create(messageGuids)))
		);
		for (String messageGuid : messageGuids) {

			var attachmentText = LoremIpsum.getInstance().getParagraphs(10, 20);
			var attachment = V1ReplyFilesTestFactory.createBuilder()
					.guid(UUID.randomUUID())
					.name(LoremIpsum.getInstance().getName() + ".txt")
					.build();
			// Stub individual response for message (MessageExchangeApi::getMessage)
			postfachFacadeMockServer.stubFor(get(urlPathTemplate("/MessageExchange/v1/Receive/{messageId}"))
					.withPathParam("messageId", equalTo(messageGuid))
					.willReturn(okJsonObj(V1ReplyMessageTestFactory.create()
							.guid(UUID.fromString(messageGuid))
							.files(List.of(attachment))
					))
			);
			// Stub attachment response (MessageExchangeApi::getMessageAttachment)
			postfachFacadeMockServer.stubFor(get(urlPathTemplate("/MessageExchange/v1/Receive/{messageId}/Attachment/{attachmentId}"))
					.withPathParam("messageId", equalTo(messageGuid))
					.withPathParam("attachmentId", equalTo(attachment.getGuid().toString()))
					.willReturn(ok()
							.withHeader("Content-Type", "application/octet-stream")
							.withBody(attachmentText.getBytes())
					)
			);
		}
	}

	private ResponseDefinitionBuilder okJsonObj(final Object body) {
		return okJson(JsonUtil.toJson(body));
	}

	@DisplayName("should delete message")
	@Test
	@SneakyThrows
	void shouldDeleteMessage() {
		var messageId = "00000000-0000-0000-0000-000000000000";
		// Stub delete message response (MessageExchangeApi::deleteMessage)
		postfachFacadeMockServer.stubFor(delete(urlPathTemplate("/MessageExchange/v1/Delete/{messageId}"))
				.willReturn(ok())
		);

		osiPostfachRemoteService.deleteMessage(messageId);

		postfachFacadeMockServer.verify(
				exactly(1),
				deleteRequestedFor(urlPathTemplate("/MessageExchange/v1/Delete/{messageId}"))
						.withPathParam("messageId", equalTo(messageId))
		);
	}
}
