package de.ozgcloud.nachrichten.postfach.osiv2;

import static de.ozgcloud.nachrichten.postfach.osiv2.config.Osi2PostfachProperties.*;
import static org.assertj.core.api.Assertions.*;

import java.util.Arrays;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.condition.EnabledIfEnvironmentVariable;
import org.junit.jupiter.api.extension.RegisterExtension;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.DynamicPropertyRegistry;
import org.springframework.test.context.DynamicPropertySource;

import de.ozgcloud.nachrichten.postfach.PostfachMessageCode;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.StringBasedIdentifier;
import de.ozgcloud.nachrichten.postfach.osiv2.attachment.Osi2AttachmentFileService;
import de.ozgcloud.nachrichten.postfach.osiv2.exception.Osi2PostfachException;
import de.ozgcloud.nachrichten.postfach.osiv2.extension.AttachmentExampleUploadUtil;
import de.ozgcloud.nachrichten.postfach.osiv2.extension.VorgangManagerServerExtension;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.PostfachAddressTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.PostfachNachrichtTestFactory;
import lombok.SneakyThrows;

@SpringBootTest(classes = TestApplication.class)
@ActiveProfiles({ "stage", "itcase" })
@EnabledIfEnvironmentVariable(named = "SH_STAGE_CLIENT_SECRET", matches = ".+")
class OsiPostfachRemoteServiceRemoteITCase {

	@Autowired
	private OsiPostfachRemoteService osiPostfachRemoteService;

	@RegisterExtension
	static final VorgangManagerServerExtension VORGANG_MANAGER_SERVER_EXTENSION = new VorgangManagerServerExtension();

	@DynamicPropertySource
	static void dynamicProperties(DynamicPropertyRegistry registry) {
		registry.add(
				"ozgcloud.osiv2.auth.client-secret",
				() -> System.getenv("SH_STAGE_CLIENT_SECRET")
		);
		registry.add(
				"ozgcloud.osiv2.proxy.enabled",
				() -> HttpProxyConfig.fromEnv().map(v -> "true").orElse("false")
		);
		registry.add(
				"ozgcloud.osiv2.proxy.host",
				() -> HttpProxyConfig.fromEnv().map(HttpProxyConfig::host).orElse("")
		);
		registry.add(
				"ozgcloud.osiv2.proxy.port",
				() -> HttpProxyConfig.fromEnv().map(HttpProxyConfig::port).orElse("")
		);
		registry.add("grpc.client." + GRPC_FILE_MANAGER_NAME + ".address", VORGANG_MANAGER_SERVER_EXTENSION::getVorgangManagerAddress);
		registry.add("grpc.client." + GRPC_FILE_MANAGER_NAME + ".negotiationType", () -> "PLAINTEXT");
	}

	@Autowired
	protected Osi2AttachmentFileService osi2AttachmentFileService;

	private PostfachNachricht createNachricht(String... attachmentIds) {
		return PostfachNachrichtTestFactory.createBuilder()
				.replyOption(PostfachNachricht.ReplyOption.POSSIBLE)
				.attachments(Arrays.stream(attachmentIds).toList())
				.postfachAddress(PostfachAddressTestFactory.createBuilder()
						.identifier(StringBasedIdentifier.builder()
								.postfachId("49b5a7e2-5e60-4baf-8ccf-1f5b94b570f3")
								.build())
						.build())
				.build();
	}

	@DisplayName("should not fail sending message")
	@Test
	@SneakyThrows
	void shouldNotFailSendingMessage() {
		var nachricht = createNachricht();

		assertThatCode(() -> osiPostfachRemoteService.sendMessage(nachricht))
				.doesNotThrowAnyException();
	}

	@DisplayName("should not fail sending message with attachment")
	@Test
	void shouldNotFailSendingMessageWithAttachment() {
		var textFileId = AttachmentExampleUploadUtil.uploadTextFile(osi2AttachmentFileService);
		var nachricht = createNachricht(textFileId);

		assertThatCode(() -> osiPostfachRemoteService.sendMessage(nachricht))
				.doesNotThrowAnyException();
	}

	@DisplayName("should throw postfach exception with unknown postfach id error code")
	@Test
	void shouldThrowPostfachExceptionWithUnknownPostfachIdErrorCode() {
		var nachrichtWithUnknownPostfachId = PostfachNachrichtTestFactory.create();

		assertThatThrownBy(() -> osiPostfachRemoteService.sendMessage(nachrichtWithUnknownPostfachId))
				.isInstanceOf(Osi2PostfachException.class)
				.hasFieldOrPropertyWithValue("messageCode", PostfachMessageCode.SEND_FAILED_UNKNOWN_POSTFACH_ID_MESSAGE_CODE);
	}

	@DisplayName("should receive messages")
	@Test
	void shouldReceiveMessages() {
		var messages = osiPostfachRemoteService.getAllMessages().toList();

		assertThat(messages)
				.isNotEmpty()
				.extracting(PostfachNachricht::getMessageId)
				.doesNotContainNull();
	}

	@DisplayName("should delete message")
	@Test
	void shouldDeleteMessage() {
		assertThatCode(() -> osiPostfachRemoteService.deleteMessage("2cec3eac-66d2-4de0-bc6b-652b8e985ceb"))
				.doesNotThrowAnyException();
	}
}
