package de.ozgcloud.nachrichten.postfach.osiv2;

import static de.ozgcloud.nachrichten.postfach.osiv2.factory.MessageExchangeReceiveMessagesResponseTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.stream.Stream;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.nachrichten.postfach.PostfachMessageCode;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.osiv2.exception.Osi2ExceptionHandler;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.PostfachNachrichtTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.transfer.Osi2PostfachService;

class OsiPostfachRemoteServiceTest {

	@InjectMocks
	@Spy
	private OsiPostfachRemoteService osiPostfachRemoteService;

	@Mock
	private Osi2PostfachService osi2PostfachService;

	@Mock
	private Osi2ExceptionHandler exceptionHandler;

	private final PostfachNachricht nachricht1 = PostfachNachrichtTestFactory.createBuilder()
			.subject("Nachricht 1")
			.build();
	private final PostfachNachricht nachricht2 = PostfachNachrichtTestFactory.createBuilder()
			.subject("Nachricht 2")
			.build();

	private final PostfachMessageCode postfachMessageCode = PostfachMessageCode.PROCESS_FAILED_MESSAGE_CODE;

	@DisplayName("send message")
	@Nested
	class TestSendOsi2Message {

		@DisplayName("should call send message")
		@Test
		void shouldCallSendMessage() {
			osi2PostfachService.sendMessage(nachricht1);

			verify(osi2PostfachService).sendMessage(nachricht1);
		}

		@DisplayName("should throw osi postfach exception on runtime exception")
		@Test
		void shouldThrowOsiPostfachExceptionOnRuntimeException() {
			var runtimeException = createRuntimeException();
			doThrow(runtimeException).when(osi2PostfachService).sendMessage(nachricht1);
			when(exceptionHandler.deriveMessageCode(runtimeException)).thenReturn(postfachMessageCode);

			assertThatThrownBy(() -> osiPostfachRemoteService.sendMessage(nachricht1))
					.hasFieldOrPropertyWithValue("messageCode", postfachMessageCode)
					.hasCause(runtimeException);
		}

	}

	@DisplayName("get all messages")
	@Nested
	class TestGetAllMessages {

		@DisplayName("should return postfach messages")
		@Test
		void shouldReturnPostfachMessages() {
			when(osi2PostfachService.receiveMessages()).thenReturn(Stream.of(nachricht1, nachricht2));

			var result = osiPostfachRemoteService.getAllMessages();

			assertThat(result).containsExactly(nachricht1, nachricht2);
		}

		@DisplayName("should throw osi postfach exception on runtime exception")
		@Test
		void shouldThrowOsiPostfachExceptionOnRuntimeException() {
			var runtimeException = createRuntimeException();
			doThrow(runtimeException).when(osi2PostfachService).receiveMessages();
			when(exceptionHandler.deriveMessageCode(runtimeException)).thenReturn(postfachMessageCode);

			assertThatThrownBy(() -> osiPostfachRemoteService.getAllMessages())
					.hasFieldOrPropertyWithValue("messageCode", postfachMessageCode)
					.hasCause(runtimeException);
		}
	}

	@DisplayName("delete message")
	@Nested
	class TestDeleteOsi2Message {

		@DisplayName("should call deleteMessage")
		@Test
		void shouldCallDeleteMessage() {
			osiPostfachRemoteService.deleteMessage(MESSAGE_ID_1);

			verify(osi2PostfachService).deleteMessage(MESSAGE_ID_1);
		}

		@DisplayName("should throw osi postfach exception on runtime exception")
		@Test
		void shouldThrowOsiPostfachExceptionOnRuntimeException() {
			var runtimeException = createRuntimeException();
			doThrow(runtimeException).when(osi2PostfachService).deleteMessage(any());
			when(exceptionHandler.deriveMessageCode(runtimeException)).thenReturn(postfachMessageCode);

			assertThatThrownBy(() -> osiPostfachRemoteService.deleteMessage(MESSAGE_ID_1))
					.hasFieldOrPropertyWithValue("messageCode", postfachMessageCode)
					.hasCause(runtimeException);
		}
	}

	@DisplayName("get postfach type")
	@Nested
	class TestGetPostfachType {
		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var result = osiPostfachRemoteService.getPostfachType();

			assertThat(result).isEqualTo(OsiPostfachRemoteService.POSTFACH_TYPE_OSI);
		}
	}

	@DisplayName("is reply allowed")
	@Nested
	class TestIsReplyAllowed {
		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var result = osiPostfachRemoteService.isReplyAllowed();

			assertThat(result).isTrue();
		}
	}

	private RuntimeException createRuntimeException() {
		return new RuntimeException();
	}

}