package de.ozgcloud.nachrichten.postfach.osiv2.exception;

import static de.ozgcloud.nachrichten.postfach.PostfachMessageCode.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.http.HttpStatusCode;
import org.springframework.web.client.ResourceAccessException;
import org.springframework.web.client.RestClientResponseException;

class Osi2ExceptionHandlerTest {

	@InjectMocks
	@Spy
	private Osi2ExceptionHandler handler;

	@DisplayName("derive message code")
	@Nested
	class TestDeriveOsi2MessageCode {
		@Mock
		RestClientResponseException restClientResponseException;

		@DisplayName("should derive message code from rest client response exception")
		@Test
		void shouldDeriveMessageCodeFromRestClientResponseException() {
			doReturn(SERVER_CONNECTION_FAILED_MESSAGE_CODE).when(handler)
					.deriveMessageCodeFromRestClientResponseException(restClientResponseException);

			var result = handler.deriveMessageCode(restClientResponseException);

			assertThat(result).isEqualTo(SERVER_CONNECTION_FAILED_MESSAGE_CODE);
		}

		@DisplayName("should return connection failed message code")
		@Test
		void shouldReturnConnectionFailedMessageCode() {
			var result = handler.deriveMessageCode(new ResourceAccessException(""));

			assertThat(result).isEqualTo(SERVER_CONNECTION_FAILED_MESSAGE_CODE);
		}

		@DisplayName("should return unknown postfach id message code")
		@Test
		void shouldReturnUnknownPostfachIdMessageCode() {
			var result = handler.deriveMessageCode(new Osi2MailboxNotFoundException(""));

			assertThat(result).isEqualTo(SEND_FAILED_UNKNOWN_POSTFACH_ID_MESSAGE_CODE);
		}

		@DisplayName("should return processed failed error code by default")
		@Test
		void shouldReturnProcessedFailedErrorCodeByDefault() {
			var result = handler.deriveMessageCode(null);

			assertThat(result).isEqualTo(PROCESS_FAILED_MESSAGE_CODE);
		}
	}

	@DisplayName("derive message code from rest client response exception")
	@Nested
	class TestDeriveOsi2MessageCodeFromRestClientResponseException {

		@Mock
		private RestClientResponseException restClientResponseException;

		@DisplayName("should return unknown postfach id error code for 404")
		@Test
		void shouldReturnUnknownPostfachIdErrorCodeFor404() {
			when(restClientResponseException.getStatusCode()).thenReturn(HttpStatusCode.valueOf(404));

			var result = handler.deriveMessageCodeFromRestClientResponseException(restClientResponseException);

			assertThat(result).isEqualTo(SEND_FAILED_UNKNOWN_POSTFACH_ID_MESSAGE_CODE);
		}

		@DisplayName("should return process failed message code by default")
		@Test
		void shouldReturnProcessFailedMessageCodeByDefault() {
			when(restClientResponseException.getStatusCode()).thenReturn(HttpStatusCode.valueOf(500));

			var result = handler.deriveMessageCodeFromRestClientResponseException(restClientResponseException);

			assertThat(result).isEqualTo(PROCESS_FAILED_MESSAGE_CODE);
		}
	}

}