package de.ozgcloud.nachrichten.postfach.osiv2.extension;

import static de.ozgcloud.nachrichten.postfach.osiv2.factory.JsonUtil.*;

import java.io.Serializable;
import java.util.Map;

import org.testcontainers.shaded.org.bouncycastle.util.encoders.Base64;

import com.jayway.jsonpath.JsonPath;
import com.jayway.jsonpath.ReadContext;

import lombok.SneakyThrows;

public record Jwt(String token) {

	public record JwtParts(ReadContext header, ReadContext body, byte[] signature) {
	}

	@SneakyThrows
	public static JwtParts parseAuthHeaderValue(String authorizationHeaderValue) {
		return new Jwt(discardBearerPrefix(authorizationHeaderValue)).parse();

	}

	public String scope() {
		return parse().body().read("$.scope");
	}

	public Integer issuedAt() {
		return parse().body().read("$.iat", Integer.class);
	}

	public Integer expireTime() {
		return parse().body().read("$.exp", Integer.class);
	}

	public JwtParts parse() {
		var jwtParts = splitIntoHeaderAndBodyAndSignature(this.token);
		return new JwtParts(
				parseJsonPath(base64UrlDecode(jwtParts[0])),
				parseJsonPath(base64UrlDecode(jwtParts[1])),
				base64UrlDecode(jwtParts[2])
		);

	}

	private static ReadContext parseJsonPath(byte[] bytes) {
		return JsonPath.parse(new String(bytes));
	}

	private static String discardBearerPrefix(String authorizationHeaderValue) {
		return authorizationHeaderValue.substring("Bearer ".length());
	}

	private static String[] splitIntoHeaderAndBodyAndSignature(String jwt) {
		var jwtParts = jwt.split("\\.");
		if (jwtParts.length != 3) {
			throw new IllegalArgumentException("Invalid JWT token");
		}
		return jwtParts;
	}

	private static byte[] base64UrlDecode(String input) {
		// Replace URL-safe characters
		String base64 = input.replace('-', '+').replace('_', '/');
		// Add padding if necessary
		int padding = 4 - (base64.length() % 4);
		if (padding < 4) {
			base64 += "=".repeat(padding);
		}
		return Base64.decode(base64);
	}

	@SneakyThrows
	public static Jwt create(Map<String, Serializable> header, Map<String, Serializable> body, byte[] signature) {
		var headerJson = toJson(header);
		var bodyJson = toJson(body);
		var token = base64UrlEncode(headerJson) + "." + base64UrlEncode(bodyJson) + "." + base64UrlEncode(new String(signature));
		return new Jwt(token);
	}

	private static String base64UrlEncode(String input) {
		var base64 = new String(Base64.encode(input.getBytes()));
		return base64.replace('+', '-').replace('/', '_').replace("=", "");

	}

}
