package de.ozgcloud.nachrichten.postfach.osiv2.extension;

import java.time.Duration;
import java.util.Map;
import java.util.Optional;

import org.junit.jupiter.api.extension.AfterAllCallback;
import org.junit.jupiter.api.extension.BeforeAllCallback;
import org.junit.jupiter.api.extension.ExtensionContext;
import org.testcontainers.containers.MongoDBContainer;
import org.testcontainers.containers.Network;
import org.testcontainers.containers.output.OutputFrame;
import org.testcontainers.lifecycle.Startables;

import lombok.Getter;
import lombok.RequiredArgsConstructor;
import lombok.SneakyThrows;
import lombok.extern.log4j.Log4j2;

@Log4j2
@Getter
@RequiredArgsConstructor
public class VorgangManagerServerExtension implements BeforeAllCallback, AfterAllCallback {

	static final String VORGANG_MANAGER_NETWORK_ALIAS = "vorgang-manager";
	static final String MONGODB_NETWORK_ALIAS = "mongodb";

	private VorgangManagerContainer vorgangManagerContainer;
	private MongoDBContainer mongoDBContainer;

	@Override
	public void beforeAll(ExtensionContext context) {
		if (vorgangManagerContainer == null || mongoDBContainer == null) {
			setupVorgangManager();
		}
	}

	@SneakyThrows
	private void setupVorgangManager() {
		var network = Network.newNetwork();
		mongoDBContainer = new MongoDBContainer("mongo:7.0")
				.withStartupTimeout(Duration.ofMinutes(1))
				.withNetworkAliases(MONGODB_NETWORK_ALIAS)
				.withNetwork(network);
		vorgangManagerContainer = new VorgangManagerContainer()
				.withEnv(Map.of(
						"spring_profiles_active", "local",
						"ozgcloud.vorgang-manager.serviceAddress", "%s:%d".formatted(VORGANG_MANAGER_NETWORK_ALIAS, VorgangManagerContainer.PORT),
						"spring.data.mongodb.host", MONGODB_NETWORK_ALIAS,
						"spring.data.mongodb.port", "27017",
						"mongock_enabled", "false"
				))
				.withNetwork(network)
				.withLogConsumer(this::logVorgangManager)
				.withStartupTimeout(Duration.ofMinutes(2))
				.dependsOn(mongoDBContainer);
		Startables.deepStart(vorgangManagerContainer, mongoDBContainer).get();
	}

	private void logVorgangManager(OutputFrame outputFrame) {
		var stream = outputFrame.getType() == OutputFrame.OutputType.STDOUT ? System.out : System.err;
		stream.println("[vorgang-manager] " + outputFrame.getUtf8String().stripTrailing());
	}

	@Override
	public void afterAll(ExtensionContext context) {
		if (mongoDBContainer != null) {
			mongoDBContainer.stop();
			mongoDBContainer = null;
		}
		if (vorgangManagerContainer != null) {
			vorgangManagerContainer.stop();
			vorgangManagerContainer = null;
		}
	}

	public String getVorgangManagerAddress() {
		return Optional.ofNullable(vorgangManagerContainer)
				.map(VorgangManagerContainer::getAddress)
				.orElse("missing");
	}

}
