package de.ozgcloud.nachrichten.postfach.osiv2.transfer;

import static de.ozgcloud.apilib.file.OzgCloudFileTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.factory.Osi2AttachmentTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.factory.Osi2MessageTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.factory.PostfachAddressTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.factory.PostfachNachrichtTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.transfer.Osi2MessageMapper.*;
import static org.assertj.core.api.Assertions.*;

import java.util.List;
import java.util.UUID;
import java.util.stream.Stream;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mapstruct.factory.Mappers;

import de.ozgcloud.apilib.file.OzgCloudFileId;
import de.ozgcloud.apilib.file.OzgCloudFileTestFactory;
import de.ozgcloud.nachrichten.postfach.PostfachAddress;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.osiv2.OsiPostfachRemoteService;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.Osi2AttachmentTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.Osi2MessageTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.PostfachAddressTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.PostfachNachrichtTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.model.AttachmentInfo;
import de.ozgcloud.nachrichten.postfach.osiv2.model.Osi2Attachment;
import de.ozgcloud.nachrichten.postfach.osiv2.model.Osi2Message;

class Osi2MessageMapperTest {

	private final Osi2MessageMapper mapper = Mappers.getMapper(Osi2MessageMapper.class);

	@DisplayName("map PostfachNachricht")
	@Nested
	class TestMapPostfachNachricht {
		private final Osi2Message osi2Message = Osi2MessageTestFactory.create();
		private final String attachmentId1 = "attachmentId1";
		private final String attachmentId2 = "attachmentId2";

		@DisplayName("should map vorgang id")
		@Test
		void shouldMapVorgangId() {
			var result = doMapping();

			assertThat(result.getVorgangId()).isEqualTo(VORGANG_ID);
		}

		@DisplayName("should map mail body")
		@Test
		void shouldMapMailBody() {
			var result = doMapping();

			assertThat(result.getMailBody()).isEqualTo(MAIL_BODY);
		}

		@DisplayName("should map subject")
		@Test
		void shouldMapSubject() {
			var result = doMapping();

			assertThat(result.getSubject()).isEqualTo(MAIL_SUBJECT);
		}

		@DisplayName("should map reply option")
		@Test
		void shouldMapReplyOption() {
			var result = doMapping();

			assertThat(result.getReplyOption()).isEqualTo(PostfachNachricht.ReplyOption.FORBIDDEN);
		}

		@DisplayName("should map created at")
		@Test
		void shouldMapCreatedAt() {
			var result = doMapping();

			assertThat(result.getCreatedAt()).isEqualTo(osi2Message.createdAt());
		}

		@DisplayName("should map postfach address")
		@Test
		void shouldMapPostfachAddress() {
			var result = doMapping();

			assertThat(result.getPostfachAddress()).usingRecursiveComparison().isEqualTo(
					mapper.createPostfachAddressByMailboxId(osi2Message.mailboxId())
			);
		}

		@DisplayName("should map direction")
		@Test
		void shouldMapDirection() {
			var result = doMapping();

			assertThat(result.getDirection()).isEqualTo(PostfachNachricht.Direction.IN);
		}

		@DisplayName("should map attachments")
		@Test
		void shouldMapAttachments() {
			var result = doMapping();

			assertThat(result.getAttachments()).containsExactly(attachmentId1, attachmentId2);
		}

		private PostfachNachricht doMapping() {
			return mapper.mapPostfachNachricht(osi2Message, List.of(attachmentId1, attachmentId2));
		}
	}

	@DisplayName("create postfach address by mailboxId")
	@Nested
	class TestCreatePostfachAddressByMailboxId {

		@DisplayName("should map postfach address type")
		@Test
		void shouldMapPostfachAddressType() {
			var result = doMapping();

			assertThat(result.getType()).isEqualTo(POSTFACH_ADDRESS_TYPE);
		}

		@DisplayName("should map version")
		@Test
		void shouldMapVersion() {
			var result = doMapping();

			assertThat(result.getVersion()).isEqualTo(POSTFACH_ADDRESS_VERSION);
		}

		@DisplayName("should map postfachId")
		@Test
		void shouldMapPostfachId() {
			var result = doMapping();

			assertThat(result.getIdentifier().getStringRepresentation()).isEqualTo(MAILBOX_ID);
		}

		@DisplayName("should map servicekontotype")
		@Test
		void shouldMapServicekontotype() {
			var result = doMapping();

			assertThat(result.getServiceKontoType()).isEqualTo(OsiPostfachRemoteService.POSTFACH_TYPE_OSI);
		}

		private PostfachAddress doMapping() {
			return mapper.createPostfachAddressByMailboxId(MAILBOX_ID);
		}
	}

	@DisplayName("map osi2 message")
	@Nested
	class TestMapOsi2Message {

		private static final String FILE_ID2 = UUID.randomUUID().toString();
		private final Osi2Attachment attachment1 = Osi2AttachmentTestFactory.createBuilder()
				.build();
		private final Osi2Attachment attachment2 = Osi2AttachmentTestFactory.createBuilder()
				.guid(ATTACHMENT_GUID_2)
				.file(OzgCloudFileTestFactory.createBuilder()
						.id(OzgCloudFileId.from(FILE_ID2))
						.build())
				.build();
		private final PostfachNachricht nachricht = PostfachNachrichtTestFactory.createBuilder()
				.attachments(List.of(ID_STR, FILE_ID2))
				.build();

		@DisplayName("should map vorgangId")
		@Test
		void shouldMapVorgangId() {
			var result = doMapping();

			assertThat(result.vorgangId()).isEqualTo(VORGANG_ID);
		}

		@DisplayName("should map mailboxId")
		@Test
		void shouldMapMailboxId() {
			var result = doMapping();

			assertThat(result.mailboxId()).isEqualTo(MAILBOX_ID);
		}

		@DisplayName("should map messsageId")
		@Test
		void shouldMapMesssageId() {
			var result = doMapping();

			assertThat(result.messageId()).isEqualTo(MESSAGE_ID);
		}

		@DisplayName("should map createdAt")
		@Test
		void shouldMapCreatedAt() {
			var result = doMapping();

			assertThat(result.createdAt()).isEqualTo(CREATED_AT);
		}

		@DisplayName("should map subject")
		@Test
		void shouldMapSubject() {
			var result = doMapping();

			assertThat(result.subject()).isEqualTo(MAIL_SUBJECT);
		}

		@DisplayName("should map mailBody")
		@Test
		void shouldMapMailBody() {
			var result = doMapping();

			assertThat(result.mailBody()).isEqualTo(MAIL_BODY);
		}

		@DisplayName("should map replyOption")
		@Test
		void shouldMapReplyOption() {
			var result = doMapping();

			assertThat(result.replyOption()).isEqualTo(REPLY_OPTION);
		}

		@DisplayName("should map attachments")
		@Test
		void shouldMapAttachments() {
			var result = doMapping();

			assertThat(result.attachments()).containsExactly(
					mapper.mapAttachmentInfo(attachment1),
					mapper.mapAttachmentInfo(attachment2)
			);
		}

		private Osi2Message doMapping() {
			return mapper.mapOsi2Message(
					nachricht,
					MAILBOX_ID,
					List.of(attachment2, attachment1)
			);
		}
	}

	@DisplayName("map attachment info")
	@Nested
	class TestMapAttachmentInfo {

		private final Osi2Attachment attachment = Osi2AttachmentTestFactory.create();

		@DisplayName("should map guid")
		@Test
		void shouldMapGuid() {
			var result = doMapping();

			assertThat(result.guid()).isEqualTo(ATTACHMENT_GUID);
		}

		@DisplayName("should map name")
		@Test
		void shouldMapName() {
			var result = doMapping();

			assertThat(result.name()).isEqualTo(OzgCloudFileTestFactory.NAME);
		}

		@DisplayName("should map contentType")
		@Test
		void shouldMapContentType() {
			var result = doMapping();

			assertThat(result.contentType()).isEqualTo(OzgCloudFileTestFactory.CONTENT_TYPE);
		}

		@DisplayName("should map size")
		@Test
		void shouldMapSize() {
			var result = doMapping();

			assertThat(result.size()).isEqualTo(OzgCloudFileTestFactory.SIZE);
		}

		private AttachmentInfo doMapping() {
			return mapper.mapAttachmentInfo(attachment);
		}
	}

	@DisplayName("get nameIdentifier")
	@Nested
	class TestGetNameIdentifier {

		@DisplayName("should map postfach address to nameIdentifier")
		@Test
		void shouldMapPostfachAddressToNameIdentifier() {
			var result = mapper.getNameIdentifier(PostfachNachrichtTestFactory.create());

			assertThat(result).isEqualTo(NAME_IDENTIFIER);
		}

		@DisplayName("should throw on missing postfach address")
		@ParameterizedTest
		@MethodSource("badPostfachAddresses")
		void shouldThrowOnMissingPostfachAddress(PostfachAddress badPostfachAddress) {
			var nachrichtWithBadPostfachAddress = PostfachNachrichtTestFactory.createBuilder()
					.postfachAddress(badPostfachAddress)
					.build();

			assertThatThrownBy(() -> mapper.getNameIdentifier(nachrichtWithBadPostfachAddress))
					.isInstanceOf(IllegalArgumentException.class);
		}

		static Stream<Arguments> badPostfachAddresses() {
			return Stream.of(
					null,
					Arguments.of(PostfachAddressTestFactory.createBuilder()
							.identifier(null)
							.build())
			);
		}
	}

}