package de.ozgcloud.nachrichten.postfach.osiv2.transfer;

import static de.ozgcloud.apilib.file.OzgCloudFileTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.factory.MessageExchangeReceiveMessagesResponseTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.factory.Osi2MessageTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.factory.PostfachAddressTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.factory.PostfachNachrichtTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.osiv2.attachment.Osi2PersistAttachmentService;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.Osi2FileUploadTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.Osi2MessageTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.PostfachNachrichtTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.model.Osi2Attachment;
import de.ozgcloud.nachrichten.postfach.osiv2.model.Osi2Message;

class Osi2PostfachServiceTest {

	@Spy
	@InjectMocks
	private Osi2PostfachService service;

	@Mock
	private PostfachApiFacadeService postfachApiFacadeService;
	@Mock
	private Osi2QuarantineService quarantineService;
	@Mock
	private Osi2PersistAttachmentService persistAttachmentService;
	@Mock
	private Osi2MessageMapper messageMapper;

	@DisplayName("send message")
	@Nested
	class TestSendOsi2Message {
		private final PostfachNachricht nachricht = PostfachNachrichtTestFactory.create();
		private final Osi2Message osi2Message = Osi2MessageTestFactory.create();
		private final List<Osi2Attachment> uploadFiles = List.of(Osi2FileUploadTestFactory.create());

		@BeforeEach
		void mock() {
			when(messageMapper.getNameIdentifier(any())).thenReturn(NAME_IDENTIFIER);
			when(messageMapper.mapOsi2Message(any(), any(), any())).thenReturn(osi2Message);
			when(quarantineService.uploadAttachments(any())).thenReturn(uploadFiles);
			when(postfachApiFacadeService.lookupMailboxId(any())).thenReturn(MAILBOX_ID);
		}

		@DisplayName("should call getNameIdentifier")
		@Test
		void shouldCallGetNameIdentifier() {
			service.sendMessage(nachricht);

			verify(messageMapper).getNameIdentifier(nachricht);
		}

		@DisplayName("should call lookupMailboxId")
		@Test
		void shouldCallLookupMailboxId() {
			service.sendMessage(nachricht);

			verify(postfachApiFacadeService).lookupMailboxId(NAME_IDENTIFIER);
		}

		@DisplayName("should call mapOsi2Message")
		@Test
		void shouldCallMapOsi2Message() {
			service.sendMessage(nachricht);

			verify(messageMapper).mapOsi2Message(nachricht, MAILBOX_ID, uploadFiles);
		}

		@DisplayName("should call uploadAttachments")
		@Test
		void shouldCallUploadAttachments() {
			service.sendMessage(nachricht);

			verify(quarantineService).uploadAttachments(nachricht.getAttachments());
		}

		@DisplayName("should call sendMessage")
		@Test
		void shouldCallSendMessage() {
			service.sendMessage(nachricht);

			verify(postfachApiFacadeService).sendMessage(osi2Message);
		}
	}

	@DisplayName("receive messages")
	@Nested
	class TestReceiveOsi2Message {

		@DisplayName("with two pending messages")
		@Nested
		class TestWithTwoPendingMessages {

			@BeforeEach
			void mock() {
				when(postfachApiFacadeService.fetchPendingMessageIds()).thenReturn(List.of(MESSAGE_ID_1, MESSAGE_ID_2));
				doReturn(PostfachNachrichtTestFactory.createBuilder().messageId(MESSAGE_ID_1).build())
						.when(service).fetchPostfachNachricht(MESSAGE_ID_1);
				doReturn(PostfachNachrichtTestFactory.createBuilder().messageId(MESSAGE_ID_2).build())
						.when(service).fetchPostfachNachricht(MESSAGE_ID_2);
			}

			@DisplayName("should return")
			@Test
			void shouldReturn() {
				var messages = receiveMessages();

				assertThat(messages)
						.extracting(PostfachNachricht::getMessageId)
						.containsExactly(MESSAGE_ID_1, MESSAGE_ID_2);
			}

		}

		private Stream<PostfachNachricht> receiveMessages() {
			return service.receiveMessages();
		}
	}

	@DisplayName("fetch PostfachNachricht")
	@Nested
	class TestFetchPostfachNachricht {

		private final Osi2Message message = Osi2MessageTestFactory.create();
		private final PostfachNachricht postfachNachricht = PostfachNachrichtTestFactory.create();

		@BeforeEach
		void mock() {
			when(postfachApiFacadeService.fetchMessageById(any())).thenReturn(message);
			when(messageMapper.mapPostfachNachricht(any(), any())).thenReturn(postfachNachricht);
			when(persistAttachmentService.persistAttachments(any())).thenReturn(List.of(ID_STR));
		}

		@DisplayName("should call fetchMessageById")
		@Test
		void shouldCallFetchMessageById() {
			fetchPostfachNachricht();

			verify(postfachApiFacadeService).fetchMessageById(MESSAGE_ID);
		}

		@DisplayName("should call mapPostfachNachricht")
		@Test
		void shouldCallMapPostfachNachricht() {
			fetchPostfachNachricht();

			verify(messageMapper).mapPostfachNachricht(message, List.of(ID_STR));
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var result = fetchPostfachNachricht();

			assertThat(result).isEqualTo(postfachNachricht);
		}

		private PostfachNachricht fetchPostfachNachricht() {
			return service.fetchPostfachNachricht(MESSAGE_ID);
		}
	}

}