package de.ozgcloud.nachrichten.postfach.osiv2.transfer;

import static de.ozgcloud.nachrichten.postfach.osiv2.factory.Osi2FileUploadTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.model.Osi2Attachment.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.transfer.Osi2QuarantineService.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import java.util.UUID;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.mockito.Spy;

import de.ozgcloud.apilib.file.OzgCloudFile;
import de.ozgcloud.apilib.file.OzgCloudFileId;
import de.ozgcloud.apilib.file.OzgCloudFileTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.attachment.Osi2AttachmentFileService;
import de.ozgcloud.nachrichten.postfach.osiv2.exception.Osi2RuntimeException;
import de.ozgcloud.nachrichten.postfach.osiv2.exception.Osi2UploadException;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.Osi2FileUploadTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.QuarantineStatus;
import de.ozgcloud.nachrichten.postfach.osiv2.model.FileChunkInfo;
import de.ozgcloud.nachrichten.postfach.osiv2.model.Osi2Attachment;
import lombok.SneakyThrows;

class Osi2QuarantineServiceTest {

	@InjectMocks
	@Spy
	private Osi2QuarantineService service;

	@Mock
	private PostfachApiFacadeService postfachApiFacadeService;

	@Mock
	private Osi2AttachmentFileService binaryFileService;

	private final OzgCloudFile file1 = OzgCloudFileTestFactory.createBuilder()
			.id(new OzgCloudFileId(UPLOAD_FILE_ID))
			.size(1000L)
			.build();
	private final OzgCloudFile file2 = OzgCloudFileTestFactory.createBuilder()
			.size(2000L)
			.build();

	private final Osi2Attachment upload1 = Osi2FileUploadTestFactory.createBuilder()
			.file(file1)
			.build();
	private final Osi2Attachment upload2 = Osi2FileUploadTestFactory.createBuilder()
			.guid(UUID.randomUUID().toString())
			.file(file2)
			.build();
	private final Osi2Attachment upload3 = Osi2FileUploadTestFactory.createBuilder()
			.guid(UUID.randomUUID().toString())
			.build();

	private final List<Osi2Attachment> uploads = List.of(upload1, upload2, upload3);

	@DisplayName("upload attachments")
	@Nested
	class TestUploadAttachments {

		@BeforeEach
		void mock() {
			when(binaryFileService.getFileMetadataOfIds(any())).thenReturn(List.of(file1));
			doReturn(List.of(upload1)).when(service).uploadFiles(any());
		}

		@DisplayName("should call getFileMetadataOfIds")
		@Test
		void shouldCallGetFileMetadataOfIds() {
			uploadAttachments();

			verify(binaryFileService).getFileMetadataOfIds(List.of(UPLOAD_FILE_ID));
		}

		@DisplayName("should call uploadFiles")
		@Test
		void shouldCallUploadFilesToQuarantineWithLargestFileFirst() {
			uploadAttachments();

			verify(service).uploadFiles(List.of(file1));
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var result = uploadAttachments();

			assertThat(result).containsExactly(upload1);
		}

		List<Osi2Attachment> uploadAttachments() {
			return service.uploadAttachments(List.of(UPLOAD_FILE_ID));
		}
	}

	@DisplayName("upload files")
	@Nested
	class TestUploadFiles {

		@BeforeEach
		void mock() {
			doReturn(uploads).when(service).deriveSortedUploadFiles(any());
		}

		@DisplayName("without exception")
		@Nested
		class TestWithoutException {
			@BeforeEach
			void mock() {
				doNothing().when(service).tryUploadSortedFiles(any());
			}

			@DisplayName("should call deriveSortedUploadFiles")
			@Test
			void shouldCallDeriveSortedUploadFiles() {
				uploadFiles();

				verify(service).deriveSortedUploadFiles(List.of(file1));
			}

			@DisplayName("should call uploadFiles")
			@Test
			void shouldCallUploadFilesToQuarantineWithLargestFileFirst() {
				uploadFiles();

				verify(service).tryUploadSortedFiles(uploads);
			}

			@DisplayName("should return")
			@Test
			void shouldReturn() {
				var result = uploadFiles();

				assertThat(result).containsExactly(upload1, upload2, upload3);
			}
		}

		@DisplayName("with exception")
		@Nested
		class TestWithException {

			private final RuntimeException exception = new RuntimeException("test");

			@BeforeEach
			void mock() {
				doThrow(exception).when(service).tryUploadSortedFiles(any());
			}

			@DisplayName("should delete attachments")
			@Test
			void shouldDeleteAttachments() {
				try {
					uploadFiles();
				} catch (RuntimeException e) {
					// ignore
				}

				verify(service).deleteAttachments(uploads);
			}

			@DisplayName("should rethrow exception")
			@Test
			void shouldRethrowException() {
				assertThatThrownBy(TestUploadFiles.this::uploadFiles)
						.isEqualTo(exception);
			}
		}

		List<Osi2Attachment> uploadFiles() {
			return service.uploadFiles(List.of(file1));
		}
	}

	@DisplayName("derive sorted upload files")
	@Nested
	class TestDeriveSortedUploadFiles {

		@DisplayName("should return largest file first")
		@Test
		void shouldReturnLargestFileFirst() {
			var result = service.deriveSortedUploadFiles(List.of(file1, file2));

			assertThat(result)
					.extracting(Osi2Attachment::file)
					.containsExactly(file2, file1);
		}

		@DisplayName("should have random upload messageGuid")
		@Test
		void shouldHaveRandomUploadGuid() {
			var result = service.deriveSortedUploadFiles(List.of(file1, file2));

			assertThat(result)
					.extracting(Osi2Attachment::guid)
					.extracting(String::length)
					.containsExactly(36, 36);
		}
	}

	@DisplayName("try upload sorted files")
	@Nested
	class TestTryUploadSortedFiles {

		@BeforeEach
		void mock() {
			doNothing().when(service).uploadFilesToQuarantine(any());
			doNothing().when(service).waitForVirusScan(any());
		}

		@DisplayName("should call uploadFilesToQuarantine")
		@Test
		void shouldCallUploadFilesToQuarantine() {
			tryUploadSortedFiles();

			verify(service).uploadFilesToQuarantine(uploads);
		}

		@DisplayName("should call waitForVirusScan")
		@Test
		void shouldCallWaitForVirusScan() {
			tryUploadSortedFiles();

			verify(service).waitForVirusScan(uploads);
		}

		void tryUploadSortedFiles() {
			service.tryUploadSortedFiles(uploads);
		}
	}

	@DisplayName("upload files to quarantine")
	@Nested
	class TestUploadFilesToQuarantine {

		@DisplayName("should call uploadFileToQuarantine for each file")
		@Test
		void shouldCallUploadFileToQuarantineForEachFile() {
			doNothing().when(service).uploadFileToQuarantine(any());

			service.uploadFilesToQuarantine(uploads);

			verify(service).uploadFileToQuarantine(upload1);
			verify(service).uploadFileToQuarantine(upload2);
			verify(service).uploadFileToQuarantine(upload3);
		}
	}

	@DisplayName("upload file to quarantine")
	@Nested
	class TestUploadFileToQuarantine {

		@Mock
		private InputStream fileInputStream;

		@BeforeEach
		void mock() {
			when(binaryFileService.downloadFileContent(any())).thenReturn(fileInputStream);
			doNothing().when(service).uploadInputStreamToQuarantine(any(), any());
		}

		@DisplayName("should call streamFileContent")
		@Test
		void shouldCallStreamFileContent() {
			uploadFileToQuarantine();

			verify(binaryFileService).downloadFileContent(UPLOAD_FILE_ID);
		}

		@DisplayName("should call uploadInputStreamToQuarantine")
		@Test
		void shouldCallUploadInputStreamToQuarantine() {
			uploadFileToQuarantine();

			verify(service).uploadInputStreamToQuarantine(upload1, fileInputStream);
		}

		@DisplayName("should close inputStream")
		@Test
		@SneakyThrows
		void shouldCloseInputStream() {
			uploadFileToQuarantine();

			verify(fileInputStream).close();
		}

		@DisplayName("should throw OsiPostfachException if close fails with IOException")
		@Test
		@SneakyThrows
		void shouldThrowOsiPostfachException() {
			doThrow(new IOException("test")).when(fileInputStream).close();

			assertThatThrownBy(this::uploadFileToQuarantine)
					.isInstanceOf(Osi2RuntimeException.class);
		}

		void uploadFileToQuarantine() {
			service.uploadFileToQuarantine(upload1);
		}
	}

	@DisplayName("stream file chunk infos")
	@Nested
	class TestStreamFileChunkInfos {

		private final Osi2Attachment upload = Osi2FileUploadTestFactory.createBuilder()
				.file(OzgCloudFileTestFactory.createBuilder()
						.size(CHUNK_SIZE * 2)
						.build())
				.build();

		private final FileChunkInfo chunkInfo1 = FileChunkInfo.builder()
				.upload(upload)
				.chunkIndex(0)
				.build();
		private final FileChunkInfo chunkInfo2 = FileChunkInfo.builder()
				.upload(upload)
				.chunkIndex(1)
				.build();
		private final FileChunkInfo emptyChunkInfo = FileChunkInfo.builder()
				.upload(upload)
				.chunkIndex(2)
				.build();

		@DisplayName("should stream chunk infos")
		@Test
		void shouldStreamChunkInfos() {
			var result = service.streamFileChunkInfos(upload).toList();

			assertThat(result).containsExactly(chunkInfo1, chunkInfo2, emptyChunkInfo);
		}
	}

	@DisplayName("check virus scan completed")
	@Nested
	class TestCheckVirusScanCompleted {

		@DisplayName("should return true if all virus scans completed")
		@Test
		void shouldReturnTrueIfAllVirusScansCompleted() {
			doReturn(true).when(service).checkOneVirusScanCompleted(upload1);
			doReturn(true).when(service).checkOneVirusScanCompleted(upload2);
			doReturn(true).when(service).checkOneVirusScanCompleted(upload3);

			var result = service.checkVirusScanCompleted(uploads);

			assertThat(result).isTrue();
		}

		@DisplayName("should return false if one virus scan not completed")
		@Test
		void shouldReturnFalseIfOneVirusScanNotCompleted() {
			doReturn(true).when(service).checkOneVirusScanCompleted(upload1);
			doReturn(false).when(service).checkOneVirusScanCompleted(upload2);

			var result = service.checkVirusScanCompleted(uploads);

			assertThat(result).isFalse();
		}
	}

	@DisplayName("check one virus scan completed")
	@Nested
	class TestCheckOneVirusScanCompleted {
		private final Osi2Attachment upload1 = Osi2FileUploadTestFactory.create();

		@DisplayName("should call checkUploadSuccessful")
		@Test
		@SneakyThrows
		void shouldCallCheckUploadSuccessful() {
			service.checkOneVirusScanCompleted(upload1);

			verify(postfachApiFacadeService).checkUploadSuccessful(UPLOAD_GUID);
		}

		@DisplayName("should return")
		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		@SneakyThrows
		void shouldReturn(boolean value) {
			when(postfachApiFacadeService.checkUploadSuccessful(any())).thenReturn(value);

			var result = service.checkOneVirusScanCompleted(upload1);

			assertThat(result).isEqualTo(value);
		}

		@DisplayName("should throw OsiRuntimeException with file metadata")
		@Test
		@SneakyThrows
		void shouldThrowOsiRuntimeExceptionWithFileMetadata() {
			var uploadException = new Osi2UploadException(QuarantineStatus.UNSAFE);
			when(postfachApiFacadeService.checkUploadSuccessful(any())).thenThrow(uploadException);

			assertThatThrownBy(() -> service.checkOneVirusScanCompleted(upload1))
					.isInstanceOf(Osi2RuntimeException.class)
					.hasMessageContaining(upload1.getLoggableString())
					.hasCause(uploadException);
		}
	}

	@DisplayName("wait for virus scan")
	@Nested
	class TestWaitForVirusScan {

		@DisplayName("should call waitUntil")
		@Test
		void shouldCallWaitUntil() {
			doReturn(true).when(service).checkVirusScanCompleted(any());

			service.waitForVirusScan(uploads);

			verify(service).checkVirusScanCompleted(uploads);
		}

		@DisplayName("should call with polling interval and timeout")
		@Test
		void shouldCallWithPollingIntervalAndTimeout() {
			try (MockedStatic<WaitUtil> mockedStatic = mockStatic(WaitUtil.class)) {
				mockedStatic.when(() -> WaitUtil.waitUntil(any(), any(), any())).thenReturn(true);

				service.waitForVirusScan(uploads);

				mockedStatic.verify(() -> WaitUtil.waitUntil(any(), eq(POLLING_INTERVAL), eq(POLLING_TIMEOUT)));
			}
		}

		@DisplayName("should throw if scan not completed before timeout")
		@Test
		void shouldThrowIfScanNotCompletedBeforeTimeout() {
			try (MockedStatic<WaitUtil> mockedStatic = mockStatic(WaitUtil.class)) {
				mockedStatic.when(() -> WaitUtil.waitUntil(any(), any(), any())).thenReturn(false);

				assertThatThrownBy(() -> service.waitForVirusScan(uploads))
						.isInstanceOf(Osi2RuntimeException.class);
			}
		}
	}

	@DisplayName("delete attachments")
	@Nested
	class TestDeleteAttachments {
		@DisplayName("should call deleteAttachment for uploads")
		@Test
		void shouldCallDeleteAttachmentForUploads() {
			service.deleteAttachments(uploads);

			verify(postfachApiFacadeService).deleteFileUpload(upload1.guid());
			verify(postfachApiFacadeService).deleteFileUpload(upload2.guid());
			verify(postfachApiFacadeService).deleteFileUpload(upload3.guid());
		}
	}

}