package de.ozgcloud.nachrichten.postfach.osiv2.transfer;

import static de.ozgcloud.nachrichten.postfach.osiv2.factory.Osi2FileUploadTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.factory.PostfachNachrichtTestFactory.*;
import static org.assertj.core.api.Assertions.*;

import java.util.List;
import java.util.UUID;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;

import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.AttachmentInfoTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.FileChunkInfoTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.Osi2MessageTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.DomainChunkMetaData;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.MessageExchangeFiles;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.OutSendMessageRequestV2;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.V1EidasLevel;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.V1FilestorageTarget;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.V1ReplyBehavior;
import de.ozgcloud.nachrichten.postfach.osiv2.model.AttachmentInfo;
import de.ozgcloud.nachrichten.postfach.osiv2.model.FileChunkInfo;
import de.ozgcloud.nachrichten.postfach.osiv2.model.Osi2Message;

class Osi2RequestMapperTest {

	private final Osi2RequestMapper mapper = Mappers.getMapper(Osi2RequestMapper.class);

	@DisplayName("map OutSendMessageRequestV2")
	@Nested
	class TestMapOutSendOsi2MessageRequestV2 {

		private final Osi2Message osi2Message = Osi2MessageTestFactory.create();

		private final AttachmentInfo attachment1 = AttachmentInfoTestFactory.create();
		private final AttachmentInfo attachment2 = AttachmentInfoTestFactory.createBuilder()
				.guid(UPLOAD_GUID2)
				.build();

		@DisplayName("should map sequence number")
		@Test
		void shouldMapSequenceNumber() {
			var result = doMapping();

			assertThat(result.getSequencenumber()).isEqualTo(VORGANG_ID);
		}

		@DisplayName("should map subject")
		@Test
		void shouldMapSubject() {
			var result = doMapping();

			assertThat(result.getSubject()).isEqualTo(MAIL_SUBJECT);
		}

		@DisplayName("should map body")
		@Test
		void shouldMapBody() {
			var result = doMapping();

			assertThat(result.getBody()).isEqualTo(MAIL_BODY);
		}

		@DisplayName("should map origin sender to null")
		@Test
		void shouldMapOriginSenderToNull() {
			var result = doMapping();

			assertThat(result.getOriginSender()).isNull();
		}

		@DisplayName("should map forbidden reply action")
		@Test
		void shouldMapForbiddenReplyAction() {
			var result = doMapping();

			assertThat(result.getReplyAction()).isEqualTo(V1ReplyBehavior.REPLYFORBIDDEN);
		}

		@DisplayName("should map possible reply action")
		@Test
		void shouldMapPossibleReplyAction() {
			var message = Osi2MessageTestFactory.createBuilder()
					.replyOption(PostfachNachricht.ReplyOption.POSSIBLE)
					.build();

			var result = mapper.mapOutSendMessageRequestV2(message);

			assertThat(result.getReplyAction()).isEqualTo(V1ReplyBehavior.REPLYPOSSIBLE);
		}

		@DisplayName("should map mandatory reply action")
		@Test
		void shouldMapMandatoryReplyAction() {
			var message = Osi2MessageTestFactory.createBuilder()
					.replyOption(PostfachNachricht.ReplyOption.MANDATORY)
					.build();

			var result = mapper.mapOutSendMessageRequestV2(message);

			assertThat(result.getReplyAction()).isEqualTo(V1ReplyBehavior.REPLYMANDATORY);
		}

		@DisplayName("should set eidasLevel")
		@Test
		void shouldSetEidasLevel() {
			var result = doMapping();

			assertThat(result.getEidasLevel()).isEqualTo(V1EidasLevel.LOW);
		}

		@DisplayName("should set isObligatory")
		@Test
		void shouldSetIsObligatory() {
			var result = doMapping();

			assertThat(result.getIsObligatory()).isFalse();
		}

		@DisplayName("should set isHtml")
		@Test
		void shouldSetIsHtml() {
			var result = doMapping();

			assertThat(result.getIsHtml()).isFalse();
		}

		@DisplayName("should map files if empty")
		@Test
		void shouldMapFilesIfEmpty() {
			var result = doMapping();

			assertThat(result.getFiles()).isEmpty();
		}

		@DisplayName("should map files")
		@Test
		void shouldMapFiles() {
			var message = Osi2MessageTestFactory.createBuilder()
					.attachments(List.of(attachment1, attachment2))
					.build();

			var result = mapper.mapOutSendMessageRequestV2(message);

			assertThat(result.getFiles()).containsExactly(
					mapper.mapMessageExchangeFile(attachment1),
					mapper.mapMessageExchangeFile(attachment2)
			);
		}

		@DisplayName("should map references")
		@Test
		void shouldMapReferences() {
			var result = doMapping();

			assertThat(result.getReferences()).isEmpty();
		}

		private OutSendMessageRequestV2 doMapping() {
			return mapper.mapOutSendMessageRequestV2(osi2Message);
		}
	}

	@DisplayName("map message exchange file")
	@Nested
	class TestMapOsi2MessageExchangeFile {
		@DisplayName("should map messageGuid")
		@Test
		void shouldMapGuid() {
			var result = mapFile();

			assertThat(result.getGuid()).isEqualTo(UUID.fromString(UPLOAD_GUID));
		}

		@DisplayName("should map mimeType")
		@Test
		void shouldMapMimeType() {
			var result = mapFile();

			assertThat(result.getMimeType()).isEqualTo(UPLOAD_CONTENT_TYPE);
		}

		@DisplayName("should map name")
		@Test
		void shouldMapName() {
			var result = mapFile();

			assertThat(result.getName()).isEqualTo(UPLOAD_NAME);
		}

		@DisplayName("should map size")
		@Test
		void shouldMapSize() {
			var result = mapFile();

			assertThat(result.getSize()).isEqualTo(UPLOAD_SIZE);
		}

		private MessageExchangeFiles mapFile() {
			return mapper.mapMessageExchangeFile(AttachmentInfoTestFactory.create());
		}
	}

	@DisplayName("map DomainChunkMetaData")
	@Nested
	class TestMapDomainChunkMetaData {

		private final FileChunkInfo chunkInfo = FileChunkInfoTestFactory.create();

		@DisplayName("should map upload messageGuid")
		@Test
		void shouldMapUploadGuid() {
			var result = doMapping();

			assertThat(result.getUploadUid()).isEqualTo(UUID.fromString(UPLOAD_GUID));
		}

		@DisplayName("should map file name")
		@Test
		void shouldMapFileName() {
			var result = doMapping();

			assertThat(result.getFileName()).isEqualTo(UPLOAD_NAME);
		}

		@DisplayName("should map content type")
		@Test
		void shouldMapContentType() {
			var result = doMapping();

			assertThat(result.getContentType()).isEqualTo(UPLOAD_CONTENT_TYPE);
		}

		@DisplayName("should map chunk index")
		@Test
		void shouldMapChunkIndex() {
			var result = doMapping();

			assertThat(result.getChunkIndex()).isEqualTo(FileChunkInfoTestFactory.CHUNK_INDEX);
		}

		@DisplayName("should map total chunks")
		@Test
		void shouldMapTotalChunks() {
			var result = doMapping();

			assertThat(result.getTotalChunks()).isEqualTo((int) NUMBER_OF_CHUNKS);
		}

		@DisplayName("should map total file size")
		@Test
		void shouldMapTotalFileSize() {
			var result = doMapping();

			assertThat(result.getTotalFileSize()).isEqualTo(UPLOAD_SIZE);
		}

		@DisplayName("should set target to unspecified")
		@Test
		void shouldSetTargetToUnspecified() {
			var result = doMapping();

			assertThat(result.getTarget()).isEqualTo(V1FilestorageTarget.UNSPECIFIED);
		}

		private DomainChunkMetaData doMapping() {
			return mapper.mapDomainChunkMetaData(chunkInfo);
		}
	}

}