package de.ozgcloud.nachrichten.postfach.osiv2.transfer;

import static de.ozgcloud.nachrichten.postfach.osiv2.factory.QuarantineFileResultTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.factory.V1ReplyFilesTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.factory.V1ReplyMessageTestFactory.*;
import static org.assertj.core.api.Assertions.*;

import java.util.List;
import java.util.stream.Stream;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.junit.jupiter.params.provider.ValueSource;
import org.mapstruct.factory.Mappers;
import org.mockito.InjectMocks;

import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.osiv2.exception.Osi2MailboxNotFoundException;
import de.ozgcloud.nachrichten.postfach.osiv2.exception.Osi2RuntimeException;
import de.ozgcloud.nachrichten.postfach.osiv2.exception.Osi2UploadException;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.Osi2FileUploadTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.QuarantineFileResultTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.ResponsesMailboxDirectoryResponseTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.V1ReplyFilesTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.V1ReplyMessageTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.QuarantineStatus;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.ResponsesMailboxDirectoryResponse;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.V1ReplyBehavior;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.V1ReplyFiles;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.V1ReplyMessage;
import de.ozgcloud.nachrichten.postfach.osiv2.model.AttachmentInfo;
import de.ozgcloud.nachrichten.postfach.osiv2.model.Osi2Message;
import lombok.SneakyThrows;

class Osi2ResponseMapperTest {

	@InjectMocks
	private Osi2ResponseMapper mapper = Mappers.getMapper(Osi2ResponseMapper.class);
	private final V1ReplyMessage message = V1ReplyMessageTestFactory.create();

	@DisplayName("toMessage")
	@Nested
	class TestToMessage {

		@Test
		void shouldMapVorgangId() {
			var result = doMapping();

			assertThat(result.vorgangId()).isEqualTo(SEQUENCE_NUMMER);
		}

		@Test
		void shouldMapPostfachAddress() {
			var result = doMapping();

			assertThat(result.mailboxId()).isEqualTo(MESSAGE_BOX_ID);
		}

		@Test
		void shouldMapCreatedAt() {
			var result = doMapping();

			assertThat(result.createdAt()).isEqualTo(RESPONSE_TIME);
		}

		@Test
		void shouldMapSubject() {
			var result = doMapping();

			assertThat(result.subject()).isEqualTo(SUBJECT);
		}

		@Test
		void shouldMapNullBodyToEmptyString() {
			var noBodyMessage = V1ReplyMessageTestFactory.createBuilder()
					.body(null)
					.build();

			var result = mapper.toMessage(noBodyMessage);

			assertThat(result.mailBody()).isEmpty();
		}

		@DisplayName("should map modified HTML body if HTML message")
		@Test
		void shouldMapModifiedHtmlBodyIfHtmlMessage() {
			var htmlMessage = V1ReplyMessageTestFactory.createBuilder()
					.body(HTML_REPLY_BODY)
					.isHtml(true)
					.build();

			var result = mapper.toMessage(htmlMessage);

			assertThat(result.mailBody()).isEqualTo(REPLY_BODY);
		}

		@DisplayName("should map unmodified body if not HTML message")
		@ParameterizedTest
		@ValueSource(strings = { REPLY_BODY, HTML_REPLY_BODY })
		void shouldMapUnmodifiedBodyIfNotHtmlMessage(String body) {
			var htmlMessage = V1ReplyMessageTestFactory.createBuilder()
					.body(body)
					.isHtml(false)
					.build();

			var result = mapper.toMessage(htmlMessage);

			assertThat(result.mailBody()).isEqualTo(body);
		}

		static Stream<Arguments> replyOptionValues() {
			return Stream.of(
					Arguments.of(V1ReplyBehavior.REPLYPOSSIBLE, PostfachNachricht.ReplyOption.POSSIBLE),
					Arguments.of(V1ReplyBehavior.REPLYMANDATORY, PostfachNachricht.ReplyOption.MANDATORY),
					Arguments.of(V1ReplyBehavior.REPLYFORBIDDEN, PostfachNachricht.ReplyOption.FORBIDDEN),
					Arguments.of(null, PostfachNachricht.ReplyOption.FORBIDDEN)
			);
		}

		@ParameterizedTest
		@MethodSource("replyOptionValues")
		void shouldMapReplyOption(V1ReplyBehavior replyAction, PostfachNachricht.ReplyOption expected) {
			var replyActionMessage = V1ReplyMessageTestFactory.createBuilder()
					.replyAction(replyAction)
					.isHtml(false)
					.build();

			var result = mapper.toMessage(replyActionMessage);

			assertThat(result.replyOption()).isEqualTo(expected);
		}

		@DisplayName("should map messageId")
		@Test
		void shouldMapMessageId() {
			var result = doMapping();

			assertThat(result.messageId()).isEqualTo(MESSAGE_ID);
		}

		@DisplayName("should map attachments")
		@Test
		void shouldMapAttachments() {
			assert message.getFiles() != null;
			var expectedAttachments = message.getFiles()
					.stream()
					.map(file -> mapper.toAttachmentInfo(file))
					.toList();

			var result = doMapping();

			assertThat(result.attachments()).usingRecursiveComparison().isEqualTo(expectedAttachments);
		}

		@DisplayName("should map null attachments to empty list")
		@Test
		void shouldMapNullAttachmentsToEmptyList() {
			var messageWithoutAttachments = V1ReplyMessageTestFactory.createBuilder()
					.files(null)
					.build();

			var result = mapper.toMessage(messageWithoutAttachments);

			assertThat(result.attachments()).isEmpty();
		}

		@DisplayName("should not fail if all fields are null")
		@Test
		void shouldNotFailIfAllFieldsAreNull() {
			var nullMessage = new V1ReplyMessage();

			assertThatCode(() -> mapper.toMessage(nullMessage))
					.doesNotThrowAnyException();
		}

		@DisplayName("should not fail if all fields are null and isHtml")
		@Test
		void shouldNotFailIfAllFieldsAreNullAndIsHtml() {
			var nullMessage = new V1ReplyMessage()
					.isHtml(true);

			assertThatCode(() -> mapper.toMessage(nullMessage))
					.doesNotThrowAnyException();
		}

		private Osi2Message doMapping() {
			return mapper.toMessage(message);
		}
	}

	@DisplayName("to AttachmentInfo")
	@Nested
	class TestToAttachmentInfo {

		private final V1ReplyFiles file = V1ReplyFilesTestFactory.create();

		@DisplayName("should map guid")
		@Test
		void shouldMapGuid() {
			var result = doMapping();

			assertThat(result.guid()).isEqualTo(ATTACHMENT_GUID);
		}

		@DisplayName("should map name")
		@Test
		void shouldMapName() {
			var result = doMapping();

			assertThat(result.name()).isEqualTo(ATTACHMENT_NAME);
		}

		@DisplayName("should map size")
		@Test
		void shouldMapSize() {
			var result = doMapping();

			assertThat(result.size()).isEqualTo(ATTACHMENT_SIZE);
		}

		@DisplayName("should map content type")
		@Test
		void shouldMapContentType() {
			var result = doMapping();

			assertThat(result.contentType()).isEqualTo(ATTACHMENT_CONTENT_TYPE);
		}

		private AttachmentInfo doMapping() {
			return mapper.toAttachmentInfo(file);
		}
	}

	@DisplayName("is safe")
	@Nested
	class TestIsSafe {
		@DisplayName("should throw for bad end status")
		@ParameterizedTest
		@ValueSource(strings = { "None", "Unsafe", "Corrupt", "Missing" })
		void shouldThrowForBadEndStatus(String status) {
			var uploadStatus = QuarantineStatus.fromValue(status);

			assertThatThrownBy(() -> mapper.isSafe(uploadStatus))
					.isInstanceOf(Osi2UploadException.class)
					.hasMessage(status);
		}

		@DisplayName("should return false for unfinished status")
		@ParameterizedTest
		@ValueSource(strings = { "Legacy", "Commited" })
		@SneakyThrows
		void shouldReturnFalseForUnfinishedStatus(String status) {
			var uploadStatus = QuarantineStatus.fromValue(status);

			var result = mapper.isSafe(uploadStatus);

			assertThat(result).isFalse();
		}

		@DisplayName("should return true for safe status")
		@Test
		@SneakyThrows
		void shouldReturnTrueForSafeStatus() {
			var result = mapper.isSafe(QuarantineStatus.SAFE);

			assertThat(result).isTrue();
		}
	}

	@DisplayName("check chunk upload success")
	@Nested
	class TestCheckChunkUploadSuccess {

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var quarantineFileResult = QuarantineFileResultTestFactory.create();

			mapper.checkChunkUploadSuccess(quarantineFileResult);
		}

		@DisplayName("should throw on failed upload")
		@Test
		void shouldThrowOnFailedUpload() {
			var quarantineFileResult = QuarantineFileResultTestFactory.createBuilder()
					.success(false)
					.error(ERROR_STRING)
					.fileUid(Osi2FileUploadTestFactory.UPLOAD_GUID)
					.build();

			assertThatThrownBy(() -> mapper.checkChunkUploadSuccess(quarantineFileResult))
					.isInstanceOf(Osi2RuntimeException.class)
					.hasMessageContaining(ERROR_STRING)
					.hasMessageContaining(Osi2FileUploadTestFactory.UPLOAD_GUID);
		}
	}

	@DisplayName("get mailboxId by tenant or first")
	@Nested
	class TestGetMailboxIdByTenantOrFirst {

		private final ResponsesMailboxDirectoryResponse response = ResponsesMailboxDirectoryResponseTestFactory.create();

		@DisplayName("should throw with null response")
		@Test
		void shouldThrowWithNullResponse() {

			assertThatThrownBy(() -> mapper.getMailboxIdByTenantOrFirst(null, ""))
					.isInstanceOf(Osi2MailboxNotFoundException.class);
		}

		@DisplayName("should throw with missing index")
		@Test
		void shouldThrowWithMissingIndex() {
			var responseWithoutMailboxIds = ResponsesMailboxDirectoryResponseTestFactory.createBuilder()
					.mailBoxIndex(List.of())
					.build();

			assertThatThrownBy(
					() -> mapper.getMailboxIdByTenantOrFirst(responseWithoutMailboxIds, ResponsesMailboxDirectoryResponseTestFactory.TENANT_SH))
					.isInstanceOf(Osi2MailboxNotFoundException.class);
		}

		@DisplayName("should throw with missing mailboxes")
		@Test
		void shouldThrowWithMissingMailboxes() {
			var responseWithoutMailboxIds = ResponsesMailboxDirectoryResponseTestFactory.createBuilder()
					.build();

			assertThatThrownBy(
					() -> mapper.getMailboxIdByTenantOrFirst(responseWithoutMailboxIds, ResponsesMailboxDirectoryResponseTestFactory.TENANT_SH))
					.isInstanceOf(Osi2MailboxNotFoundException.class);
		}

		@DisplayName("should return mailboxId for SH")
		@Test
		void shouldReturnMailboxIdForSH() {
			var result = mapper.getMailboxIdByTenantOrFirst(response, ResponsesMailboxDirectoryResponseTestFactory.TENANT_SH);

			assertThat(result).isEqualTo(ResponsesMailboxDirectoryResponseTestFactory.MAILBOX_GUID_SH);
		}

		@DisplayName("should return mailboxId for HH")
		@Test
		void shouldReturnMailboxIdForHh() {
			var result = mapper.getMailboxIdByTenantOrFirst(response, ResponsesMailboxDirectoryResponseTestFactory.TENANT_HH);

			assertThat(result).isEqualTo(ResponsesMailboxDirectoryResponseTestFactory.MAILBOX_GUID_HH);
		}

		@DisplayName("should return first mailboxId for unknown tenant")
		@Test
		void shouldReturnFirstMailboxIdForUnknownTenant() {
			var result = mapper.getMailboxIdByTenantOrFirst(response, "unknown");

			assertThat(result).isEqualTo(ResponsesMailboxDirectoryResponseTestFactory.MAILBOX_GUID_SH);
		}

	}
}
