package de.ozgcloud.nachrichten.postfach.osiv2.transfer;

import static de.ozgcloud.nachrichten.postfach.osiv2.factory.MessageExchangeReceiveMessagesResponseTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.factory.Osi2FileUploadTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.factory.Osi2MessageTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.factory.PostfachAddressTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.factory.ResponsesMailboxDirectoryResponseTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.osiv2.transfer.Osi2RequestMapper.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.UUID;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.core.io.AbstractResource;

import de.ozgcloud.nachrichten.postfach.osiv2.config.Osi2PostfachProperties;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.FileChunkInfoTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.MessageExchangeReceiveMessagesResponseTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.Osi2MessageTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.ResponsesMailboxDirectoryResponseTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.V1ReplyMessageTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.api.MailboxDirectoryApi;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.api.MessageExchangeApi;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.api.QuarantineApi;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.DomainChunkMetaData;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.MessageExchangeDeleteMessageResponse;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.MessageExchangeReceiveMessagesResponse;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.MessageExchangeSendMessageResponse;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.OutSendMessageRequestV2;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.QuarantineStatus;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.ResponsesMailboxDirectoryResponse;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.model.V1ReplyMessage;
import de.ozgcloud.nachrichten.postfach.osiv2.model.FileChunkInfo;
import de.ozgcloud.nachrichten.postfach.osiv2.model.Osi2Message;
import lombok.SneakyThrows;

class PostfachApiFacadeServiceTest {

	@InjectMocks
	@Spy
	PostfachApiFacadeService service;

	@Mock
	MessageExchangeApi messageExchangeApi;

	@Mock
	QuarantineApi quarantineApi;

	@Mock
	MailboxDirectoryApi mailboxDirectoryApi;

	@Mock
	Osi2RequestMapper osi2RequestMapper;

	@Mock
	Osi2ResponseMapper osi2ResponseMapper;

	@Mock
	Osi2PostfachProperties.ApiConfiguration apiConfiguration;

	@DisplayName("send message")
	@Nested
	class TestSendOsi2Message {

		@Mock
		OutSendMessageRequestV2 outSendMessageRequestV2;

		@Mock
		MessageExchangeSendMessageResponse messageExchangeSendMessageResponse;

		private final Osi2Message message = Osi2MessageTestFactory.create();

		@BeforeEach
		void mock() {
			when(osi2RequestMapper.mapOutSendMessageRequestV2(any())).thenReturn(outSendMessageRequestV2);
			when(messageExchangeApi.sendMessage(any(), any())).thenReturn(messageExchangeSendMessageResponse);
		}

		@DisplayName("should call mapOutSendMessageRequestV2")
		@Test
		void shouldCallMapOutSendMessageRequestV2() {
			service.sendMessage(message);

			verify(osi2RequestMapper).mapOutSendMessageRequestV2(message);
		}

		@DisplayName("should call sendMessage")
		@Test
		void shouldCallSendMessage() {
			service.sendMessage(message);

			verify(messageExchangeApi).sendMessage(MAILBOX_ID, outSendMessageRequestV2);
		}

	}

	@DisplayName("fetch pending message ids")
	@Nested
	class TestFetchPendingOsi2MessageIds {

		@DisplayName("with two pending messages")
		@Nested
		class TestWithTwoPendingMessages {
			private final MessageExchangeReceiveMessagesResponse response = MessageExchangeReceiveMessagesResponseTestFactory.create(
					MESSAGE_ID_1, MESSAGE_ID_2
			);

			@BeforeEach
			void mock() {
				when(messageExchangeApi.receiveMessages(anyInt(), anyInt())).thenReturn(response);
				when(osi2ResponseMapper.toMessageIds(any())).thenReturn(List.of(MESSAGE_ID_1, MESSAGE_ID_2));
			}

			@DisplayName("should return")
			@Test
			void shouldReturn() {
				var messageIds = fetchMessageIds();

				assertThat(messageIds).containsExactly(MESSAGE_ID_1, MESSAGE_ID_2);
			}

			@DisplayName("should call mapper")
			@Test
			void shouldCallMapper() {
				fetchMessageIds();

				verify(osi2ResponseMapper).toMessageIds(response);
			}

			@DisplayName("should call receiveMessages api method")
			@Test
			void shouldCallReceiveMessagesApiMethod() {
				fetchMessageIds();

				verify(messageExchangeApi).receiveMessages(MAX_NUMBER_RECEIVED_MESSAGES, 0);
			}

		}

		@DisplayName("with no pending messages")
		@Nested
		class TestWithNoPendingMessages {
			private final MessageExchangeReceiveMessagesResponse emptyResponse = MessageExchangeReceiveMessagesResponseTestFactory.create();

			@BeforeEach
			void mock() {
				when(messageExchangeApi.receiveMessages(anyInt(), anyInt())).thenReturn(emptyResponse);
			}

			@DisplayName("should return")
			@Test
			void shouldReturn() {
				var messages = fetchMessageIds();

				assertThat(messages).isEmpty();
			}

		}

		private List<String> fetchMessageIds() {
			return service.fetchPendingMessageIds();
		}

	}

	@DisplayName("fetch message by id")
	@Nested
	class TestFetchMessageById {

		private final V1ReplyMessage replyMessage = V1ReplyMessageTestFactory.create();
		private final Osi2Message message = Osi2MessageTestFactory.create();

		@Test
		void shouldCallGetMessage() {
			when(messageExchangeApi.getMessage(any())).thenReturn(replyMessage);

			service.fetchMessageById(MESSAGE_ID_1);

			verify(messageExchangeApi).getMessage(UUID.fromString(MESSAGE_ID_1));
		}

		@Test
		void shouldCallResponseMapper() {
			when(messageExchangeApi.getMessage(any())).thenReturn(replyMessage);
			when(osi2ResponseMapper.toMessage(any())).thenReturn(Osi2MessageTestFactory.create());

			service.fetchMessageById(MESSAGE_ID_1);

			verify(osi2ResponseMapper).toMessage(replyMessage);
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			when(messageExchangeApi.getMessage(any())).thenReturn(replyMessage);
			when(osi2ResponseMapper.toMessage(any())).thenReturn(message);

			var result = service.fetchMessageById(MESSAGE_ID_1);

			assertThat(result).isEqualTo(message);
		}
	}

	@DisplayName("upload chunk")
	@Nested
	class TestUploadChunk {
		@Mock
		AbstractResource chunkResource;

		@Mock
		DomainChunkMetaData domainChunkMetaData;

		private static final String TENANT = "tenant";
		private static final String NAME_IDENTIFIER = "nameIdentifier";

		private final FileChunkInfo info = FileChunkInfoTestFactory.create();

		@BeforeEach
		void mock() {
			when(apiConfiguration.getTenant()).thenReturn(TENANT);
			when(apiConfiguration.getNameIdentifier()).thenReturn(NAME_IDENTIFIER);
			when(osi2RequestMapper.mapDomainChunkMetaData(any())).thenReturn(domainChunkMetaData);
		}

		@DisplayName("should call mapDomainChunkMetadata")
		@Test
		void shouldCallMapDomainChunkMetadata() {
			service.uploadChunk(info, chunkResource);

			verify(osi2RequestMapper).mapDomainChunkMetaData(info);
		}

		@DisplayName("should call uploadChunk")
		@Test
		void shouldCallUploadChunk() {
			service.uploadChunk(info, chunkResource);

			verify(quarantineApi).uploadChunk(domainChunkMetaData, TENANT, NAME_IDENTIFIER, chunkResource);
		}
	}

	@DisplayName("check upload successful")
	@Nested
	class TestCheckUploadSuccessful {

		private final QuarantineStatus uploadStatus = QuarantineStatus.SAFE;

		@BeforeEach
		@SneakyThrows
		void mock() {
			when(osi2ResponseMapper.isSafe(any())).thenReturn(true);
			when(quarantineApi.getUploadStatus(any())).thenReturn(uploadStatus);
		}

		@DisplayName("should call getUploadStatus")
		@Test
		@SneakyThrows
		void shouldCallGetUploadStatus() {
			service.checkUploadSuccessful(MESSAGE_ID_1);

			verify(quarantineApi).getUploadStatus(UUID.fromString(MESSAGE_ID_1));
		}

		@DisplayName("should call isSafe")
		@Test
		@SneakyThrows
		void shouldCallIsSafe() {
			service.checkUploadSuccessful(MESSAGE_ID_1);

			verify(osi2ResponseMapper).isSafe(uploadStatus);
		}

		@DisplayName("should return true")
		@Test
		@SneakyThrows
		void shouldReturnTrue() {
			var result = service.checkUploadSuccessful(MESSAGE_ID_1);

			assertThat(result).isTrue();
		}

	}

	@DisplayName("delete message")
	@Nested
	class TestDeleteOsi2Message {
		@Mock
		MessageExchangeDeleteMessageResponse replyMessage;

		@Test
		void shouldCallDeleteMessage() {
			when(messageExchangeApi.deleteMessage(any())).thenReturn(replyMessage);

			service.deleteMessage(MESSAGE_ID_1);

			verify(messageExchangeApi).deleteMessage(UUID.fromString(MESSAGE_ID_1));
		}
	}

	@DisplayName("delete file upload")
	@Nested
	class TestDeleteFileUpload {
		@DisplayName("should call deleteUpload")
		@Test
		void shouldCallDeleteUpload() {
			service.deleteFileUpload(MESSAGE_ID_1);

			verify(quarantineApi).deleteUpload(UUID.fromString(MESSAGE_ID_1));
		}
	}

	@DisplayName("download attachment")
	@Nested
	class TestDownloadAttachment {
		@Mock
		AbstractResource resource;

		@BeforeEach
		void mock() {
			when(messageExchangeApi.getMessageAttachment(any(), any())).thenReturn(resource);
		}

		@DisplayName("should call getMessageAttachment")
		@Test
		void shouldCallGetMessageAttachment() {
			service.downloadAttachment(MESSAGE_ID_1, UPLOAD_GUID);

			verify(messageExchangeApi).getMessageAttachment(UUID.fromString(MESSAGE_ID_1), UUID.fromString(UPLOAD_GUID));
		}

		@DisplayName("should return resource")
		@Test
		void shouldReturnResource() {
			var result = service.downloadAttachment(MESSAGE_ID_1, UPLOAD_GUID);

			assertThat(result).isEqualTo(resource);
		}
	}

	@DisplayName("lookup mailboxId")
	@Nested
	class TestLookupMailboxId {

		private final ResponsesMailboxDirectoryResponse response = ResponsesMailboxDirectoryResponseTestFactory.create();

		@BeforeEach
		void mock() {
			when(osi2ResponseMapper.getMailboxIdByTenantOrFirst(any(), any())).thenReturn(MAILBOX_ID);
			when(mailboxDirectoryApi.lookupMailboxIds(any())).thenReturn(response);
			when(apiConfiguration.getTenant()).thenReturn(TENANT_SH);
		}

		@DisplayName("should call lookupMailboxIds")
		@Test
		void shouldCallLookupMailboxIds() {
			lookupMailboxId();

			verify(mailboxDirectoryApi).lookupMailboxIds(List.of(UUID.fromString(NAME_IDENTIFIER)));
		}

		@DisplayName("should call getMailBoxIdByTenantOrFirst")
		@Test
		void shouldCallGetMailBoxIdByTenantOrFirst() {
			lookupMailboxId();

			verify(osi2ResponseMapper).getMailboxIdByTenantOrFirst(response, TENANT_SH);
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var result = lookupMailboxId();

			assertThat(result).isEqualTo(MAILBOX_ID);
		}

		private String lookupMailboxId() {
			return service.lookupMailboxId(NAME_IDENTIFIER);
		}
	}

}