package de.ozgcloud.nachrichten.postfach.osiv2.transfer;

import static de.ozgcloud.nachrichten.postfach.osiv2.transfer.WaitUtil.*;
import static org.assertj.core.api.Assertions.*;
import static org.awaitility.Awaitility.*;

import java.time.Duration;
import java.util.concurrent.CompletableFuture;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;

import lombok.SneakyThrows;

class WaitUtilTest {

	@DisplayName("wait until")
	@Nested
	class TestWaitUntil {
		@DisplayName("should return true if check condition is eventually true")
		@Test
		void shouldReturnTrueIfCheckConditionIsEventuallyTrue() {
			var delayFuture = createDelayedFuture(100);

			var conditionFuture = CompletableFuture.supplyAsync(() -> waitForDelayedCondition(delayFuture));

			await()
					.between(Duration.ofMillis(100), Duration.ofMillis(500))
					.until(conditionFuture::isDone);
			assertThat(conditionFuture.getNow(false)).isTrue();
		}

		@DisplayName("should return false if check condition is not true before timeout")
		@Test
		void shouldReturnFalseIfCheckConditionIsNotTrueBeforeTimeout() {
			var delayCondition = createDelayedFuture(1000);

			var conditionFuture = CompletableFuture.supplyAsync(() -> waitForDelayedCondition(delayCondition));

			await()
					.between(Duration.ofMillis(500), Duration.ofMillis(1000))
					.until(conditionFuture::isDone);
			assertThat(conditionFuture.getNow(true)).isFalse();
		}

		@DisplayName("should rethrow runtime exceptions of condition")
		@Test
		void shouldRethrowRuntimeExceptionsOfCondition() {
			var exception = new RuntimeException("test");

			var conditionFuture = CompletableFuture.supplyAsync(() ->
					waitUntil(() -> {
						throw exception;
					}, Duration.ofMillis(50), Duration.ofMillis(500)));

			await()
					.atMost(Duration.ofMillis(500))
					.until(conditionFuture::isDone);
			assertThatThrownBy(() -> conditionFuture.getNow(false))
					.hasCause(exception);
		}

		private CompletableFuture<Boolean> createDelayedFuture(long milliseconds) {
			return CompletableFuture.supplyAsync(() -> {
				try {
					Thread.sleep(milliseconds);
					return true;
				} catch (InterruptedException e) {
					throw new RuntimeException(e);
				}
			});
		}

		@SneakyThrows
		private boolean waitForDelayedCondition(CompletableFuture<Boolean> future) {
			return waitUntil(future::isDone, Duration.ofMillis(50), Duration.ofMillis(500));
		}
	}

}